import time
from functools import reduce
from itertools import product
from operator import or_, and_
from warnings import warn

import networkx
import numpy as np
import logging
from pathlib import Path
from typing import Union, List, Tuple, Type, Dict, Optional, Set
import pandas as pd
import re

import networkx as nx
import py2neo
from collections import Counter

from astropy.io import fits
from tqdm import tqdm

from weaveio.address import Address
from weaveio.basequery.common import AmbiguousPathError
from weaveio.basequery.handler import Handler, defaultdict
from weaveio.basequery.tree import BranchHandler
from weaveio.basequery.actions import TraversalPath
from weaveio.graph import Graph
from weaveio.writequery import Unwind
from weaveio.hierarchy import Multiple, Hierarchy, Indexed, Graphable, One2One
from weaveio.file import File, HDU

CONSTRAINT_FAILURE = re.compile(r"already exists with label `(?P<label>[^`]+)` and property "
                                r"`(?P<idname>[^`]+)` = (?P<idvalue>[^`]+)$", flags=re.IGNORECASE)

def process_neo4j_error(data: 'Data', file: File, msg):
    matches = CONSTRAINT_FAILURE.findall(msg)
    if not len(matches):
        return  # cannot help
    label, idname, idvalue = matches[0]
    # get the node properties that already exist
    extant = data.graph.neograph.evaluate(f'MATCH (n:{label} {{{idname}: {idvalue}}}) RETURN properties(n)')
    fname = data.graph.neograph.evaluate(f'MATCH (n:{label} {{{idname}: {idvalue}}})-[*]->(f:File) return f.fname limit 1')
    idvalue = idvalue.strip("'").strip('"')
    file.data = data
    obj = [i for i in data.hierarchies if i.__name__ == label][0]
    instance_list = getattr(file, obj.plural_name)
    new = {}
    if not isinstance(instance_list, (list, tuple)):  # has an unwind table object
        new_idvalue = instance_list.identifier
        if isinstance(new_idvalue, Unwind):
            # find the index in the table and get the properties
            filt = (new_idvalue.data == idvalue).iloc[:, 0]
            for k in extant.keys():
                if k == 'id':
                    k = idname
                value = getattr(instance_list, k, None)
                if isinstance(value, Unwind):
                    table = value.data.where(pd.notnull(value.data), 'NaN')
                    new[k] = str(table[k][filt].values[0])
                else:
                    new[k] = str(value)
        else:
            # if the identifier of this object is not looping through a table, we cant proceed
            return
    else:  # is a list of non-table things
        found = [i for i in instance_list if i.identifier == idvalue][0]
        for k in extant.keys():
            value = getattr(found, k, None)
            new[k] = value
    comparison = pd.concat([pd.Series(extant, name='extant'), pd.Series(new, name='to_add')], axis=1)
    filt = comparison.extant != comparison.to_add
    filt &= ~comparison.isnull().all(axis=1)
    where_different = comparison[filt]
    logging.exception(f"The node (:{label} {{{idname}: {idvalue}}}) tried to be created twice with different properties.")
    logging.exception(f"{where_different}")
    logging.exception(f"filenames: {fname}, {file.fname}")


def get_all_subclasses(cls: Type[Graphable]) -> List[Type[Graphable]]:
    all_subclasses = []
    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))
    return all_subclasses


def get_all_class_bases(cls: Type[Graphable]) -> List[Type[Graphable]]:
    new = []
    for b in cls.__bases__:
        if b is Graphable or not issubclass(b, Graphable):
            continue
        new.append(b)
        new += get_all_class_bases(b)
    return new



def find_children_of(parent):
    hierarchies = get_all_subclasses(Hierarchy)
    children = set()
    for h in hierarchies:
        if len(h.parents):
            if any(p is parent if isinstance(p, type) else p.node is parent for p in h.parents):
                children.add(h)
    return children


class IndirectAccessError(Exception):
    pass


class MultiplicityError(Exception):
    pass


def shared_base_class(*classes):
    if len(classes):
        all_classes = list(reduce(and_, [set(get_all_class_bases(cls)+[cls]) for cls in classes]))
        all_classes.sort(key=lambda x: len(get_all_class_bases(x)), reverse=True)
        if all_classes:
            return all_classes[0]
    return Hierarchy


class Data:
    filetypes = []

    @staticmethod
    def hierarchies_from_filetype(*filetype: Type[File]) -> Set[Type[Hierarchy]]:
        hierarchies = set()
        all_hierarchies = set(get_all_subclasses(Hierarchy))
        todo = set(filetype)
        while todo:
            thing = todo.pop()
            if isinstance(thing, Multiple):
                thing = thing.node
            if thing in hierarchies:
                continue  # if already done
            if thing.is_template:
                if not issubclass(thing, File):
                    todo.update(get_all_subclasses(thing))
                continue  # but dont add it directly yet
            todo.update(thing.produces)
            todo.update(thing.hdus.values())
            todo.update(thing.parents)
            bases = get_all_class_bases(thing)
            bases.append(thing)
            todo.update([i for i in all_hierarchies if any(b.singular_name in i.belongs_to for b in bases) and not i.is_template])  # children
            hierarchies.add(thing)
            hierarchies.update(get_all_class_bases(thing))
        return hierarchies


    def __init__(self, rootdir: Union[Path, str] = '/beegfs/weave/weaveio/',
                 host: str = '127.0.0.1', port=7687, write=False,
                 password='weavepassword', user='weaveuser'):
        self.branch_handler = BranchHandler()
        self.handler = Handler(self)
        self.host = host
        self.port = port
        self.write_allowed = write
        self._graph = None
        self.password = password
        self.user = user
        self.filelists = {}
        self.rootdir = Path(rootdir)
        self.relation_graphs = []
        for i, f in enumerate(self.filetypes):
            fs = self.filetypes[:i+1]
            self.relation_graphs.append(self.make_relation_graph(self.hierarchies_from_filetype(*fs), fs))
        self.relation_graph = self.relation_graphs[-1]
        self.traversal_graph = self.make_traversal_graph()
        self.hierarchies = list(self.relation_graph.nodes)
        self.hierarchies += list({template for h in self.hierarchies for template in get_all_class_bases(h) if template.is_template})
        self.class_hierarchies = {h.__name__: h for h in self.hierarchies}
        self.singular_hierarchies = {h.singular_name: h for h in self.hierarchies}  # type: Dict[str, Type[Hierarchy]]
        self.plural_hierarchies = {h.plural_name: h for h in self.hierarchies if h.plural_name != 'graphables'}
        self.factor_hierarchies = defaultdict(list)
        for h in self.hierarchies:
            for f in getattr(h, 'products_and_factors', []):
                self.factor_hierarchies[f.lower()].append(h)
            if h.idname is not None:
                self.factor_hierarchies[h.idname].append(h)
        self.factor_hierarchies = dict(self.factor_hierarchies)  # make sure we always get keyerrors when necessary!
        self.factors = set(self.factor_hierarchies.keys())
        self.plural_factors =  {f.lower() + 's': f.lower() for f in self.factors}
        self.singular_factors = {f.lower() : f.lower() for f in self.factors}
        self.singular_idnames = {h.idname: h for h in self.hierarchies if h.idname is not None}
        self.plural_idnames = {k+'s': v for k,v in self.singular_idnames.items()}


    def write(self, collision_manager='track&flag'):
        if self.write_allowed:
            return self.graph.write(collision_manager)
        raise IOError(f"You have not allowed write operations in this instance of data (write=False)")

    def is_unique_factor(self, name):
        return len(self.factor_hierarchies[name]) == 1

    @property
    def graph(self):
        if self._graph is None:
            d = {}
            if self.password is not None:
                d['password'] = self.password
            if self.user is not None:
                d['user'] = self.user
            self._graph = Graph(host=self.host, port=self.port, write=self.write, **d)
        return self._graph

    def make_traversal_graph(self):
        graph = self.relation_graph.reverse(copy=True)
        return graph

    def make_relation_graph(self, hierarchies, filetypes):
        reference = hierarchies
        hierarchies = hierarchies.copy()
        graph = nx.DiGraph()
        while hierarchies:
            hierarchy = hierarchies.pop()
            if hierarchy.is_template:
                continue
            graph.add_node(hierarchy)
            for parent in set(hierarchy.parents + hierarchy.produces):
                multiplicity = isinstance(parent, Multiple) and not isinstance(parent, One2One)
                is_one2one = isinstance(parent, One2One)
                if multiplicity:
                    if parent.maxnumber == parent.minnumber:
                        number = parent.maxnumber
                        numberlabel = f'={number}'
                    else:
                        number = None
                        if (parent.minnumber is None or parent.minnumber == 0) and parent.maxnumber is None:
                            numberlabel = 'any'
                        elif (parent.minnumber is None or parent.minnumber == 0) and parent.maxnumber is not None:
                            numberlabel = f'<= {parent.maxnumber}'
                        elif (parent.minnumber is not None and parent.minnumber > 0) and parent.maxnumber is None:
                            numberlabel = f'>={parent.minnumber}'
                        else:
                            numberlabel = f'{parent.minnumber} - {parent.maxnumber}'
                else:
                    number = 1
                    numberlabel = f'={number}'
                if isinstance(parent, Multiple):
                    parent = parent.node
                reverse = parent in hierarchy.produces
                data = dict(multiplicity=multiplicity, number=number, label=numberlabel)
                if parent.is_template:
                    to_do = set(get_all_subclasses(parent))
                    data['oneway'] = True
                    data['label'] += ' (oneway)'
                else:
                    to_do = {parent}
                for p in to_do:
                    if p not in reference:
                        continue
                    if reverse:
                        graph.add_edge(hierarchy, p, **data, real=True)
                        if is_one2one:
                            graph.add_edge(p, hierarchy, **data, real=False)
                    else:
                        graph.add_edge(p, hierarchy, **data, real=True)
                        if is_one2one:
                            graph.add_edge(hierarchy, p, **data, real=False)
                        elif p.singular_name in hierarchy.belongs_to or any(s.singular_name in hierarchy.belongs_to for s in  get_all_subclasses(p)):
                            graph.add_edge(hierarchy, p, multiplicity=True, number=None, label='any', real=False)
        bunch = [i for i in filetypes]
        for filetype in filetypes:
            bunch += list(nx.ancestors(graph, filetype))
            bunch += list(nx.descendants(graph, filetype))
        temp = nx.DiGraph(nx.subgraph(graph, bunch))
        view = nx.subgraph_view(temp, lambda n: not n.is_template)
        return nx.DiGraph(view)

    def make_constraints_cypher(self):
        return {hierarchy: hierarchy.make_schema() for hierarchy in self.hierarchies}

    def apply_constraints(self):
        if not self.write_allowed:
            raise IOError(f"Writing is not allowed")
        templates = []
        equivalencies = []
        for hier, q in tqdm(self.make_constraints_cypher().items(), desc='applying constraints'):
            if q is None:
                templates.append(hier)
            else:
                try:
                    self.graph.neograph.run(q)
                except py2neo.ClientError as e:
                    if '[Schema.EquivalentSchemaRuleAlreadyExists]' in str(e):
                       equivalencies.append(hier)
                       templates.append(hier)
        if len(templates):
            print(f'No index/constraint was made for {templates}')
        if len(equivalencies):
            print(f'EquivalentSchemaRuleAlreadyExists for {equivalencies}')

    def drop_all_constraints(self):
        if not self.write_allowed:
            raise IOError(f"Writing is not allowed")
        self.graph.neograph.run('CALL apoc.schema.assert({},{},true) YIELD label, key RETURN *')

    def get_extant_files(self):
        return self.graph.execute("MATCH (f:File) RETURN DISTINCT f.fname").to_series(dtype=str).values.tolist()

    def raise_collisions(self):
        """
        returns the properties that would have been overwritten in nodes and relationships.
        """
        node_collisions = self.graph.execute("MATCH (c: _Collision) return c { .*}").to_data_frame()
        rel_collisions = self.graph.execute("MATCH ()-[c: _Collision]-() return c { .*}").to_data_frame()
        return node_collisions, rel_collisions

    def read_files(self, *paths: Union[Path, str], collision_manager='ignore', batch_size=None, halt_on_error=False) -> pd.DataFrame:
        """
        Read in the files given in `paths` to the database.
        `collision_manager` is the method with which the database deals with overwriting data.
        Values of `collision_manager` can be {'ignore', 'overwrite', 'track&flag'}.
        track&flag will have the same behaviour as ignore but places the overlapping data in its own node for later retrieval.
        :return
            statistics dataframe
        """
        batches = []
        for path in paths:
            path = Path(path)
            matches = [f for f in self.filetypes if f.match_file(self.rootdir, path.relative_to(self.rootdir), self.graph)]
            if len(matches) > 1:
                raise ValueError(f"{path} matches more than 1 file type: {matches} with `{[m.match_pattern for m in matches]}`")
            filetype = matches[0]
            filetype_batch_size = filetype.recommended_batchsize if batch_size is None else batch_size
            slices = filetype.get_batches(path, filetype_batch_size)
            batches += [(filetype, path.relative_to(self.rootdir), slc) for slc in slices]
        elapsed_times = []
        stats = []
        timestamps = []
        bar = tqdm(batches)
        for filetype, fname, slc in bar:
            bar.set_description(f'{fname}[{slc.start}:{slc.stop}]')
            with self.write(collision_manager) as query:
                filetype.read(self.rootdir, fname, slc)
            cypher, params = query.render_query()
            start = time.time()
            try:
                results = self.graph.execute(cypher, **params)
                stats.append(results.stats())
                timestamp = results.evaluate()
                if timestamp is None:
                    warn(f"This query terminated early due to an empty input table/data. "
                         f"Adjust your `.read` method to allow for empty tables/data")
                timestamps.append(timestamp)
                elapsed_times.append(time.time() - start)
            except py2neo.database.work.ClientError as e:
                logging.exception('ClientError:', exc_info=True)
                if halt_on_error:
                    raise e
                print(e)
        if len(batches):
            df = pd.DataFrame(stats)
            df['timestamp'] = timestamps
            df['elapsed_time'] = elapsed_times
            _, df['fname'], slcs = zip(*batches)
            df['batch_start'], df['batch_end'] = zip(*[(i.start, i.stop) for i in slcs])
        else:
            df = pd.DataFrame(columns=['timestamp', 'elapsed_time', 'fname', 'batch_start', 'batch_end'])
        return df.set_index(['fname', 'batch_start', 'batch_end'])

    def find_files(self, *filetype_names, skip_extant_files=True):
        filelist = []
        if len(filetype_names) == 0:
            filetypes = self.filetypes
        else:
            filetypes = [f for f in self.filetypes if f.singular_name in filetype_names or f.plural_name in filetype_names]
        for filetype in filetypes:
            filelist += [i for i in self.rootdir.rglob(filetype.match_pattern)]
        if skip_extant_files:
            extant_fnames = self.get_extant_files() if skip_extant_files else []
            filtered_filelist = [i for i in filelist if str(i.relative_to(self.rootdir)) not in extant_fnames]
        else:
            filtered_filelist = filelist
        diff = len(filelist) - len(filtered_filelist)
        if diff:
            print(f'Skipping {diff} extant files (use skip_extant_files=False to go over them again)')
        return filtered_filelist

    def read_directory(self, *filetype_names, collision_manager='ignore', skip_extant_files=True, halt_on_error=False) -> pd.DataFrame:
        filtered_filelist = self.find_files(*filetype_names, skip_extant_files=skip_extant_files)
        return self.read_files(*filtered_filelist, collision_manager=collision_manager, halt_on_error=halt_on_error)

    def _validate_one_required(self, hierarchy_name):
        hierarchy = self.singular_hierarchies[hierarchy_name]
        parents = [h for h in hierarchy.parents]
        qs = []
        for parent in parents:
            if isinstance(parent, Multiple):
                mn, mx = parent.minnumber, parent.maxnumber
                b = parent.node.__name__
            else:
                mn, mx = 1, 1
                b = parent.__name__
            mn = 0 if mn is None else mn
            mx = 9999999 if mx is None else mx
            a = hierarchy.__name__
            q = f"""
            MATCH (n:{a})
            WITH n, SIZE([(n)<-[]-(m:{b}) | m ])  AS nodeCount
            WHERE NOT (nodeCount >= {mn} AND nodeCount <= {mx})
            RETURN "{a}", "{b}", {mn} as mn, {mx} as mx, n.id, nodeCount
            """
            qs.append(q)
        if not len(parents):
            qs = [f"""
            MATCH (n:{hierarchy.__name__})
            WITH n, SIZE([(n)<-[:IS_REQUIRED_BY]-(m) | m ])  AS nodeCount
            WHERE nodeCount > 0
            RETURN "{hierarchy.__name__}", "none", 0 as mn, 0 as mx, n.id, nodeCount
            """]
        dfs = []
        for q in qs:
            dfs.append(self.graph.neograph.run(q).to_data_frame())
        df = pd.concat(dfs)
        return df

    def _validate_no_duplicate_relation_ordering(self):
        q = """
        MATCH (a)-[r1]->(b)<-[r2]-(a)
        WHERE TYPE(r1) = TYPE(r2) AND r1.order <> r2.order
        WITH a, b, apoc.coll.union(COLLECT(r1), COLLECT(r2))[1..] AS rs
        RETURN DISTINCT labels(a), a.id, labels(b), b.id, count(rs)+1
        """
        return self.graph.neograph.run(q).to_data_frame()

    def _validate_no_duplicate_relationships(self):
        q = """
        MATCH (a)-[r1]->(b)<-[r2]-(a)
        WHERE TYPE(r1) = TYPE(r2) AND PROPERTIES(r1) = PROPERTIES(r2)
        WITH a, b, apoc.coll.union(COLLECT(r1), COLLECT(r2))[1..] AS rs
        RETURN DISTINCT labels(a), a.id, labels(b), b.id, count(rs)+1
        """
        return self.graph.neograph.run(q).to_data_frame()

    def validate(self):
        duplicates = self._validate_no_duplicate_relationships()
        print(f'There are {len(duplicates)} duplicate relations')
        if len(duplicates):
            print(duplicates)
        duplicates = self._validate_no_duplicate_relation_ordering()
        print(f'There are {len(duplicates)} relations with different orderings')
        if len(duplicates):
            print(duplicates)
        schema_violations = []
        for h in tqdm(list(self.singular_hierarchies.keys())):
            schema_violations.append(self._validate_one_required(h))
        schema_violations = pd.concat(schema_violations)
        print(f'There are {len(schema_violations)} violations of expected relationship number')
        if len(schema_violations):
            print(schema_violations)
        return duplicates, schema_violations

    def find_factor_paths(self, starting_point: Type[Hierarchy], factor_name: str,
                          plural: bool) -> Tuple[Dict[Type[Hierarchy], Set[TraversalPath]], Type[Hierarchy]]:
        """
        1. Identify all hierarchies that contain the factor under plural constraint
        2. Get paths to those hierarchies with the plural constraint
        3. Discard the hierarchies which don't have paths
        4.
        """
        if factor_name in starting_point.products_and_factors:
            return {starting_point: set()}, starting_point
        possible = {c for c in get_all_subclasses(Hierarchy) if factor_name in c.products_and_factors and not c.is_template}
        pathset = set()
        for p in possible:
            try:
                paths, ends, _, _ = self.find_hierarchy_paths(starting_point, p, plural)
                for path, end in zip(paths, ends):
                    pathset.add((path, end))
            except nx.NetworkXNoPath:
                pass
        if len(pathset) == 0:
            raise nx.NetworkXNoPath(f'There are no paths from {starting_point} to {factor_name} with the constraint of plural={plural}')
        paths, ends = zip(*pathset)
        if not plural and len(paths) > 1:
            lengths = map(len, paths)
            min_length = min(lengths)
            paths, ends = zip(*[(p, e) for p, e in zip(paths, ends) if len(p) == min_length])
            if len(paths) > 1:
                raise AmbiguousPathError(f"There is more than one {factor_name} with the same distance away from {starting_point}")
        shared = shared_base_class(*ends)
        if factor_name not in shared.products_and_factors:
            raise AmbiguousPathError(f"{starting_point}.{factor_name} refers to multiple objects ({ends}) which have no consistent shared parent")
        pathdict = defaultdict(set)
        for e, p in zip(ends, paths):
            pathdict[e].add(p)
        return pathdict, shared

    @staticmethod
    def shortest_path_without_oneway_violation(graph: nx.Graph, a, b, cutoff=50):
        """Iterate over shortest paths until one without reusing a one-way path is found"""
        for count, path in enumerate(nx.shortest_simple_paths(graph, a, b)):
            if count == cutoff:
                raise nx.NetworkXNoPath(f'No viable path between {a} and {b} (exploration timed out)')
            oneway_node = None
            for x, y in zip(path[:-1], path[1:]):
                oneway = graph.edges[(x, y)].get('oneway', False)  # is it oneway?
                if oneway:
                    if x == oneway_node:  # violated
                        break
                    else:
                        oneway_node = y
            else:
                return path
        else:
            raise nx.NetworkXNoPath(f'No viable path between {a} and {b}')

    def _find_restricted_path(self, traversal_graph: nx.DiGraph, a: Type[Hierarchy], b: Type[Hierarchy],
                              plural: bool) -> Tuple[TraversalPath, List[bool], nx.DiGraph]:
        if plural:
            graph = traversal_graph
        else:
            graph = nx.subgraph_view(traversal_graph, filter_edge=lambda x, y: not traversal_graph.edges[(x, y)]['multiplicity'])
        try:
            travel_path = self.shortest_path_without_oneway_violation(graph, a, b)
        except nx.NetworkXNoPath:
            path = self.shortest_path_without_oneway_violation(graph, b, a)
            travel_path = path[::-1]
        multiplicity = []
        number = []
        _direction = []
        one_way = []
        for x, y in zip(travel_path[:-1], travel_path[1:]):
            try:
                edge = graph.edges[(y, x)]
            except KeyError:
                multiplicity.append(True)
                number.append(None)
                one_way.append(False)
            else:
                multiplicity.append(edge['multiplicity'])
                number.append(edge['number'])
                one_way.append(edge.get('oneway', False))
        direction = []
        # now reverse the arrow direction if that direction is not real
        for x, y in zip(travel_path[:-1], travel_path[1:]):
            try:
                real = traversal_graph.edges[(x, y)].get('real', False)
            except KeyError:
                real = not traversal_graph.edges[(y, x)].get('real', False)
            arrow = '->' if real else '<-'
            direction.append(arrow)
        total_path = []
        for i, node in enumerate(travel_path[1:]):
            total_path.append(direction[i])
            total_path.append(node.__name__)
        return TraversalPath(*total_path), one_way, graph

    def _find_hierarchy_path(self, a: Type[Hierarchy], b: Type[Hierarchy], plural: bool) -> Tuple[TraversalPath, List[bool], nx.DiGraph]:
        for i, graph in enumerate(self.relation_graphs):
            try:
                return self._find_restricted_path(graph, a, b, plural)
            except (nx.NetworkXNoPath, nx.NodeNotFound, KeyError) as e:
                continue
        raise nx.NetworkXNoPath(f"The is no path between {a} and {b} with a constraint of plural={plural}")

    def find_hierarchy_paths(self, a: Type[Hierarchy], b: Type[Hierarchy],
                             plural: bool) -> Tuple[List[TraversalPath], List[Type[Hierarchy]], Type[Hierarchy], Type[Hierarchy]]:
        if a.is_template:
            a = [i for i in get_all_subclasses(a) if not i.is_template]
        else:
            a = [a]
        if b.is_template:
            b = [i for i in get_all_subclasses(b) if not i.is_template]
        else:
            b = [b]
        paths = set()
        for ai, bi in product(a, b):
            try:
                path, oneways, graph = self._find_hierarchy_path(ai, bi, plural)
                if sum(oneways) and len(oneways) > 1:  # if >1 there are more paths, if 1 then you are directly connected (so dont explore)
                    if sum(oneways) > 1:
                        raise NotImplementedError(f"Resolving paths with more than 1 oneway is not yet implemented: {path.repr_path}")
                    elif oneways[-1] != True:
                        raise NotImplementedError(f"Resolving paths with the oneway not at the end is not yet implemented: {path.repr_path}")
                    else:
                        just_before = set(nx.descendants(graph, ai)) & set(graph.predecessors(bi))
                        for before in just_before:
                            path, _, _ = self._find_restricted_path(graph, ai, before, plural)
                            paths.add((TraversalPath(*path._path, '->', bi.__name__), bi))
                else:
                    paths.add((path, bi))
            except nx.NetworkXNoPath:
                pass
        if len(paths) == 0:
            raise nx.NetworkXNoPath(f'There are no paths from {a} to {b} with the constraint of plural={plural}')
        paths, ends = zip(*paths)
        if not plural and len(paths) > 1:
            lengths = map(len, paths)
            min_length = min(lengths)
            paths, ends = zip(*[(p, e) for p, e in zip(paths, ends) if len(p) == min_length])
            if len(paths) > 1:
                raise AmbiguousPathError(f"There is more than one {b} with the same distance away from {a}")
        return list(paths), list(ends), shared_base_class(*a), shared_base_class(*ends)

    def is_factor_name(self, name):
        try:
            name = self.singular_name(name)
            return self.is_singular_factor(name) or self.is_singular_idname(name)
        except KeyError:
            return False

    def is_singular_idname(self, value):
        return value.split('.')[-1] in self.singular_idnames

    def is_plural_idname(self, value):
        return value.split('.')[-1] in self.plural_idnames

    def is_plural_factor(self, value):
        return value.split('.')[-1] in self.plural_factors

    def is_singular_factor(self, value):
        return value.split('.')[-1] in self.singular_factors

    def plural_name(self, name):
        split = name.split('.')
        before, name = '.'.join(split[:-1]), split[-1]
        if self.is_plural_name(name):
            return name
        if name in self.singular_idnames:
            return name + 's'
        else:
            try:
                return before + self.singular_factors[name] + 's'
            except KeyError:
                return before + self.singular_hierarchies[name].plural_name

    def singular_name(self, name):
        split = name.split('.')
        before, name = '.'.join(split[:-1]), split[-1]
        if self.is_singular_name(name):
            return name
        if name in self.plural_idnames:
            return name[:-1]
        else:
            try:
                return before + self.plural_factors[name]
            except KeyError:
                return before + self.plural_hierarchies[name].singular_name

    def is_valid_name(self, name):
        if isinstance(name, str):
            return self.is_plural_name(name) or self.is_singular_name(name)
        return False

    def is_plural_name(self, name):
        """
        Returns True if name is a plural name of a hierarchy
        e.g. spectra is plural for Spectrum
        """
        name = name.split('.')[-1]
        return name in self.plural_hierarchies or name in self.plural_factors or name in self.plural_idnames

    def is_singular_name(self, name):
        name = name.split('.')[-1]
        return name in self.singular_hierarchies or name in self.singular_factors or name in self.singular_idnames

    def __getitem__(self, address):
        return self.handler.begin_with_heterogeneous().__getitem__(address)

    def __getattr__(self, item):
        return self.handler.begin_with_heterogeneous().__getattr__(item)

    def plot_relations(self, i=-1, show_hdus=True, fname='relations.pdf', include=None):
        from networkx.drawing.nx_agraph import to_agraph
        if not show_hdus:
            G = nx.subgraph_view(self.relation_graphs[i], lambda n: not issubclass(n, HDU))  # True to get rid of templated
        else:
            G = self.relation_graphs[i]
        if include is not None:
            include = [self.singular_hierarchies[i] for i in include]
            include_list = include.copy()
            include_list += [a for i in include for a in nx.ancestors(G, i)]
            include_list += [d for i in include for d in nx.descendants(G, i)]
            G = nx.subgraph_view(G, lambda n: n in include_list)
        A = to_agraph(G)
        A.layout('dot')
        A.draw(fname)



import xxhash


class Varname:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

def quote(x):
    """
    Return a quoted string if x is a string otherwise return x
    """
    if isinstance(x, str):
        return f"'{x}'"
    if isinstance(x, Varname):
        return x.name
    return x


def hash_pandas_dataframe(df):
    digester = xxhash.xxh64()
    for i in df.astype(str).values.ravel():
        digester.update(i)
    return digester.hexdigest()


from typing import Callable

from weaveio.basequery.common import FrozenQuery
from weaveio.basequery.dissociated import Dissociated
from weaveio.data import Data


def _template_aggregator(string_function, name, normal: Callable, item: Dissociated, wrt: FrozenQuery = None):
    if not isinstance(item, FrozenQuery):
        if wrt is not None:
            return normal(item, wrt)
        return normal(item)
    elif isinstance(wrt, Data) or wrt is None:
        branch = item.branch.handler.entry
    else:
        branch = wrt.branch
    new = branch.aggregate(string_function, item.variable, item.branch, True, name)
    return Dissociated(item.handler, new, new.action.target)


def sum(item, wrt=None):
    return _template_aggregator('sum({x})', 'sum', sum, item, wrt)


def max(item, wrt=None):
    return _template_aggregator('max({x})', 'max', max, item, wrt)


def min(item, wrt=None):
    return _template_aggregator('min({x})', 'min', min, item, wrt)


def all(item, wrt=None):
    return _template_aggregator('all(i in {x} where i)', 'all', all, item, wrt)


def any(item, wrt=None):
    return _template_aggregator('any(i in {x} where i)', 'any', any, item, wrt)


def count(item, wrt=None):
    return _template_aggregator('count({x})', 'count', len, item, wrt)



from typing import Tuple, Dict

from weaveio.data import Data
from weaveio.file import File
from weaveio.hierarchy import Hierarchy, Multiple

"""
File1<-HierarchyA<-HierarchyB<-Multiple(HierarchyC)
                             <-HierarchyD
"""

class HierarchyF(Hierarchy):
    factors = ['f_factor_a']
    idname = 'id'


class HierarchyD(Hierarchy):
    factors = ['shared_factor_name']
    idname = 'id'


class HierarchyC(Hierarchy):
    factors = ['c_factor_a', 'c_factor_b', 'shared_factor_name']
    idname = 'id'


class HierarchyB(Hierarchy):
    parents = [Multiple(HierarchyC, 2, 2), HierarchyD]
    factors = ['b_factor_a', 'b_factor_b']
    idname = 'otherid'


class HierarchyA(Hierarchy):
    parents = [HierarchyB, Multiple(HierarchyF, 1, 3)]
    factors = ['a_factor_a', 'a_factor_b']
    idname = 'id'
    version_on = ['hierarchyb']


class File1(File):
    parents = [HierarchyA]
    match_pattern = '*.fits'

    @classmethod
    def read(cls, directory, fname) -> 'File':
        fname = str(fname)
        hierarchyf = HierarchyF(id=fname, f_factor_a='a')
        hierarchyd = HierarchyD(id=fname, shared_factor_name='shared_d')
        hierarchyc1 = HierarchyC(id=fname+'1', c_factor_a='a', c_factor_b='b', shared_factor_name='shared_c1')
        hierarchyc2 = HierarchyC(id=fname+'2', c_factor_a='a', c_factor_b='b', shared_factor_name='shared_c2')
        hierarchyb = HierarchyB(otherid=fname, b_factor_b='b',  b_factor_a='a', hierarchycs=[hierarchyc1, hierarchyc2], hierarchyd=hierarchyd)
        return cls(fname, hierarchya=HierarchyA(id=fname, hierarchyb=hierarchyb, a_factor_a='a', a_factor_b='b', hierarchyfs=[hierarchyf]))


class MyDataOne2One(Data):
    filetypes = [File1]



import pandas as pd

progtemp_config = pd.DataFrame(
    {
    'mode': ['MOS']*3 + ['LIFU']*3 + ['mIFU']*3,
    'resolution': ['LowRes', 'HighRes', 'HighRes']*3,
    'red_vph': [1, 2, 2]*3,
    'blue_vph': [1, 2, 3]*3
    }, index=range(1, 10)
)



from contextlib import contextmanager
from typing import List, Callable, Union

from .base import CypherQuery, CypherVariable, Collection, CustomStatement
from .statements import Unwind, Collect, GroupBy


@contextmanager
def unwind(*args, enumerated=False):
    query = CypherQuery.get_context()  # type: CypherQuery
    unwinder = Unwind(*args, enumerated=enumerated)
    query.open_context()  # allow this context to accumulate variables
    query.add_statement(unwinder)  # append the actual statement
    if len(unwinder.passed_outputs) == 1:
        yield unwinder.passed_outputs[0]  # give back the unwound variables
    else:
        yield tuple(unwinder.passed_outputs)
    query.close_context()  # remove the variables from being referenced from now on
    query.open_contexts = [[param for param in c if param not in args] for c in query.open_contexts]
    # previous = [param for context in query.open_contexts for param in context if param not in args]
    previous = [param for context in query.open_contexts for param in context]
    query.statements.append(Collect(previous))  # allow the collections to be accessible - force


def collect(*variables: CypherVariable):
    query = CypherQuery.get_context()  # type: CypherQuery
    collector = query.statements[-1]
    if not isinstance(collector, Collect):
        raise NameError(f"You must use collect straight after a with context")
    for variable in variables:
        if variable not in query.closed_context:
            raise ValueError(f"Cannot collect a non unwound variable")
    collector = Collect(collector.previous, *variables)
    query.statements[-1] = collector
    r = [collector[variable] for variable in variables]
    query.open_contexts[-1] += r
    if len(r) == 1:
        return r[0]
    return r


def groupby(variable_list, propertyname):
    if not isinstance(variable_list, Collection):
        raise TypeError(f"{variable_list} is not a collection")
    query = CypherQuery.get_context()  # type: CypherQuery
    g = GroupBy(variable_list, propertyname)
    query.add_statement(g)
    return g.output_variables[0]


def custom(statement: Callable, inputs: List[CypherVariable] = None, returns: List[Union[CypherVariable, str]] = None,
           outputs: List[Union[CypherVariable, str]] = None):
    if returns is None:
        returns = []
    if inputs is None:
        inputs = []
    if outputs is None:
        outputs = []
    for i, o in enumerate(returns):
        if isinstance(o, str):
            returns[i] = CypherVariable(o)
        elif not isinstance(o, CypherVariable):
            raise TypeError(f"output_variables  must be str or cyphervariable")
    for i, o in enumerate(outputs):
        if isinstance(o, str):
            outputs[i] = CypherVariable(o)
        elif not isinstance(o, CypherVariable):
            raise TypeError(f"output_variables  must be str or cyphervariable")
    outputs = returns + outputs
    statement = CustomStatement(statement, inputs, outputs)
    query = CypherQuery.get_context()  # type: CypherQuery
    query.add_statement(statement)
    return returns


from textwrap import dedent
from typing import List, Dict, Union, Tuple, Optional, Iterable

from . import CypherQuery
from .base import camelcase, Varname, Statement, CypherVariable, CypherData, CypherVariableItem


def are_different(a: str, b: str) -> str:
    return f"apoc.coll.different([apoc.coll.flatten([[{a}]]), apoc.coll.flatten([[{b}]])])"


def neo4j_dictionary(d: Union[dict, CypherVariable]) -> Tuple[Union[dict, CypherVariable], List[CypherVariable]]:
    """
    If d is a cyphervariable, return it
    If not, then we submit all the individual entries as data params to neo4j
    This avoids translating the data types ourselves!
    """
    if isinstance(d, CypherVariable):
        return d, [d]
    assert all(isinstance(k, str) for k in d.keys()), "keys must be strings"
    newd = {}
    ins = []
    for k, v in d.items():
        assert v is not None
        k = Varname(k)
        if not isinstance(v, CypherVariable):
            v = CypherData(v)
        ins.append(v)
        newd[k] = v
    return newd, ins


def sanitise_variablename(v):
    return f'{str(v).replace("$", "")}'


def expand_to_cypher_dict(*collections: Union[Dict[str, CypherVariable], CypherVariable]) -> str:
    inputs = []
    for collection in collections:
        if isinstance(collection, dict):
            inputs += list(collection.values())
        elif isinstance(collection, list):
            inputs += collection
        elif isinstance(collection, CypherVariable):
            inputs.append(collection)
        else:
            raise TypeError(f"Cannot convert {collection} to a cypher input dict of form `{{x:x}}` ")
    inputs = [getattr(i, 'parent') if isinstance(i, CypherVariableItem) else i for i in inputs]
    l = set([f"{sanitise_variablename(v)}: {v}" for v in inputs])
    return ', '.join(l)


def expand_to_cypher_alias(*collections: Union[Dict[str, CypherVariable], CypherVariable],
                           prefix='$') -> str:
    inputs = []
    for collection in collections:
        if isinstance(collection, dict):
            inputs += list(collection.values())
        elif isinstance(collection, list):
            inputs += collection
        elif isinstance(collection, CypherVariable):
            inputs.append(collection)
        else:
            raise TypeError(f"Cannot convert {collection} to a cypher input dict of form `{{x:x}}` ")
    inputs = [getattr(i, 'parent') if isinstance(i, CypherVariableItem) else i for i in inputs]
    l = set([f"{prefix}{sanitise_variablename(v)} as {sanitise_variablename(v)}" for v in inputs])
    return ', '.join(l)


class MatchNode(Statement):
    keyword = 'MATCH'

    def __init__(self, labels: List[str], properties: dict):
        self.labels = [camelcase(l) for l in labels]
        self.properties, inputs = neo4j_dictionary(properties)
        self.out = CypherVariable(labels[0])
        super(MatchNode, self).__init__(inputs, [self.out])

    def to_cypher(self):
        labels = ':'.join(map(str, self.labels))
        return f"{self.keyword} ({self.out}:{labels} {self.properties})"


class MatchRelationship(Statement):
    def __init__(self, parent, child, reltype: str, properties: dict):
        self.parent = parent
        self.child = child
        self.reltype = reltype
        self.properties, inputs = neo4j_dictionary(properties)
        inputs += [self.parent, self.child]
        self.out = CypherVariable(reltype)
        super().__init__(inputs, [self.out])

    def to_cypher(self):
        reldata = f'[{self.out}:{self.reltype} {self.properties}]'
        return f"MATCH ({self.parent})-{reldata}->({self.child})"


class MatchPatternNode(Statement):
    def __init__(self, labels: List[str], properties: Dict[str, Union[str, int, float, CypherVariable]],
                 parents: List[CypherVariable], children: List[CypherVariable]):
        self.labels = [camelcase(l) for l in labels]
        self.properties, inputs = neo4j_dictionary(properties)
        self.parents = parents
        self.children = children
        self.out = CypherVariable(labels[0])
        super().__init__(inputs+parents+children, [self.out], [])

    def to_cypher(self):
        labels = ':'.join(self.labels)
        match = f'({self.out}: {labels} {self.properties})'
        wheres = ' AND '.join([f'({self.out})<--({p})' for p in self.parents] +
                            [f'({self.out})-->({c})' for c in self.children])
        if len(wheres):
            wheres = f'\nWHERE {wheres}'
        return f'WITH * OPTIONAL MATCH {match}{wheres}'


class PropertyOverlapError(Exception):
    pass


class NullPropertyError(Exception):
    pass


class CollisionManager(Statement):
    def __init__(self, out, identproperties: Dict[str, Union[str, int, float, CypherVariable]],
                 properties: Dict[str, Union[str, int, float, CypherVariable]], collision_manager='track&flag'):
        self.out = out
        self.properties, propinputs = neo4j_dictionary(properties)
        self.identproperties, identinputs = neo4j_dictionary(identproperties)
        self.validate_properties()
        self.propvar = CypherVariable('props')
        self.colliding_keys = CypherVariable('colliding_keys')
        self.value = CypherVariable('unnamed')
        inputs = propinputs + identinputs
        outputs = [self.out, self.propvar]
        if collision_manager == 'track&flag':
            outputs += [self.value, self.colliding_keys]
        elif collision_manager not in ['overwrite', 'ignore']:
            raise ValueError(f"Unknown collision_manager {collision_manager}")
        self.collision_manager = collision_manager
        super().__init__(inputs, outputs)

    def validate_properties(self):
        if any(p in self.identproperties for p in self.properties.keys()):
            raise PropertyOverlapError(f"Cannot have the same key in both properties and identproperties")
        for k, v in self.identproperties.items():
            if v != v:
                raise NullPropertyError(f"Cannot assign a nan to a node identify property")

    @property
    def on_match(self):
        if self.collision_manager == 'overwrite':
            return f"SET {self.out} += {self.propvar}   // overwrite with new colliding properties"
        return f"SET {self.out} = apoc.map.merge({self.propvar}, properties({self.out}))   // update, keeping the old colliding properties"

    @property
    def on_create(self):
        return f"SET {self.out}._dbcreated = time0, {self.out} += {self.propvar}  // setup as standard"

    @property
    def on_run(self):
        return f'SET {self.out}._dbupdated = time0  // always set updated time '

    @property
    def merge_statement(self):
        raise NotImplementedError

    @property
    def collision_record(self):
        raise NotImplementedError

    @property
    def collision_record_input(self):
        raise NotImplementedError

    @property
    def post_merge(self):
        return dedent(f"""
    ON MATCH {self.on_match}
    ON CREATE {self.on_create}
    {self.on_run}""")

    @property
    def pre_merge(self):
        return f"WITH *, {self.properties} as {self.propvar}"

    @property
    def merge_paragraph(self):
        return f"""
        {self.pre_merge}
        {self.merge_statement}
        {self.post_merge}
        """

    def to_cypher(self):
        query = self.merge_paragraph
        if self.collision_manager == 'track&flag':
            query += f"""
            WITH *, [x in apoc.coll.intersection(keys({self.propvar}), keys(properties({self.out}))) where ({self.propvar}[x] is null or {self.out}[x] is null) or {self.propvar}[x] <> {self.out}[x]] as {self.colliding_keys}
            CALL apoc.do.when(size({self.colliding_keys}) > 0, 
                "{self.collision_record} SET c = $collisions SET c._dbcreated = $time RETURN $time", 
                "RETURN $time",
                {{{self.collision_record_input}, collisions: apoc.map.fromLists({self.colliding_keys}, apoc.map.values({self.propvar}, {self.colliding_keys})), time:time0}}) yield value as {self.value}
            """
        return dedent(query)


class MergeNode(CollisionManager):
    def __init__(self, labels: List[str], identproperties: Dict[str, Union[str, int, float, CypherVariable]],
                 properties: Dict[str, Union[str, int, float, CypherVariable]], collision_manager='track&flag'):
        self.labels = [camelcase(l) for l in labels]
        out = CypherVariable(labels[0])
        super().__init__(out, identproperties, properties, collision_manager)

    @property
    def merge_statement(self):
        labels = ':'.join(map(str, self.labels))
        return f'MERGE ({self.out}: {labels} {self.identproperties})'

    @property
    def collision_record(self):
        return f"WITH $innode as innode CREATE (c:_Collision)-[:COLLIDES]->(innode)"

    @property
    def collision_record_input(self):
        return f"innode: {self.out}"


class MergeRelationship(CollisionManager):
    def __init__(self, parent, child, reltype: str, identproperties: dict, properties: dict, collision_manager='track&flag'):
        self.parent = parent
        self.child = child
        self.reltype = reltype
        out = CypherVariable(reltype)
        super().__init__(out, identproperties, properties, collision_manager)

    @property
    def merge_statement(self):
        return f'MERGE ({self.parent})-[{self.out}:{self.reltype} {self.identproperties}]->({self.child})'

    @property
    def collision_record(self):
        return f"WITH $a as a, $b as b CREATE (a)-[c:COLLIDES]->(b)"

    @property
    def collision_record_input(self):
        return f"a:{self.parent}, b:{self.child}"


class MergeDependentNode(CollisionManager):
    def __init__(self, labels: List[str], identproperties: Dict[str, Union[str, int, float, CypherVariable]],
                 properties: Dict[str, Union[str, int, float, CypherVariable]],
                 parents: List[CypherVariable],
                 reltypes: List[str],
                 relidentproperties: List[Dict[str, Union[str, int, float, CypherVariable]]],
                 relproperties: List[Dict[str, Union[str, int, float, CypherVariable]]],
                 collision_manager='track&flag'):
        if not (len(parents) == len(reltypes) == len(relproperties) == len(relidentproperties)):
            raise ValueError(f"Parents must have the same length as reltypes, relproperties, relidentproperties")
        self.labels = [camelcase(l) for l in labels]
        self.relidentproperties, relidentpropins = [], []
        self.relproperties, relpropins = [], []
        for ident, prop in zip(relidentproperties, relproperties):
            identdict, identpropins = neo4j_dictionary(ident)
            propdict, propins = neo4j_dictionary(prop)
            self.relidentproperties.append(identdict)
            self.relproperties.append(propdict)
            relidentpropins += identpropins
            relpropins += propins
        self.parents = parents
        self.outnode = CypherVariable(labels[0])
        self.relvars = [CypherVariable(reltype) for reltype in reltypes]
        self.dummyrelvars = [CypherVariable('dummy'+reltype) for reltype in reltypes]
        self.dummy = CypherVariable('dummy')
        self.reltypes = reltypes
        self.relpropsvars = [CypherVariable(f'{t}_props') for t in reltypes]
        self.colliding_rel_keys = [CypherVariable('colliding_rel_keys') for _ in reltypes]
        super().__init__(self.outnode, identproperties, properties, collision_manager)
        self.child_holder = CypherVariable('child_holder')
        self.unnamed = CypherVariable('unnamed')
        self.input_variables += parents
        self.input_variables += relidentpropins
        self.input_variables += relpropins
        self.output_variables += self.relvars
        self.output_variables += self.dummyrelvars
        self.output_variables += self.relpropsvars
        self.hidden_variables += self.colliding_rel_keys
        self.output_variables.append(self.dummy)
        self.output_variables.append(self.child_holder)
        self.output_variables.append(self.unnamed)

    def validate_properties(self):
        super(MergeDependentNode, self).validate_properties()
        for idents, props in zip(self.relidentproperties, self.relproperties):
            if any(p in idents for p in props.keys()):
                raise ValueError(f"Cannot have the same key in both properties and identproperties")
            for k, v in idents.items():
                if v != v:
                    raise NullPropertyError(f"Cannot assign a nan to a node identify property")

    @property
    def pre_merge(self):
        line = f"WITH *, {self.properties} as {self.propvar}"
        for relprop, relpropsvar in zip(self.relproperties, self.relpropsvars):
            line += f', {relprop} as {relpropsvar}'
        return line

    @property
    def merge_statement(self):
        labels = ':'.join(map(str, self.labels))
        relations = []
        for i, (parent, reltype, relidentprop, dummyrelvar) in enumerate(zip(self.parents, self.reltypes, self.relidentproperties, self.dummyrelvars)):
            rel = f'({parent})-[{dummyrelvar}:{reltype} {relidentprop}]->'
            if i == 0:
                child = f'({self.dummy}: {labels} {self.identproperties})'
            else:
                child = f'({self.dummy})'
            relations.append(rel + child)
        optional_match = 'OPTIONAL MATCH ' + ',\n'.join(relations)
        create = 'CREATE ' + ',\n'.join(relations)
        for dummy, real in zip(self.dummyrelvars, self.relvars):
            create = create.replace(f'[{dummy}:', f'[{real}:')
        create = create.replace(f'{self.dummy}', f'{self.out}')
        on_create_rel_returns = ', '.join([f'{relvar}' for relvar in self.relvars])
        on_match_rel_returns = ', '.join([f'{dummy} as {real}' for dummy, real in zip(self.dummyrelvars, self.relvars)])
        rel_expansion = expand_to_cypher_alias(self.out, *self.relvars, prefix=f'{self.child_holder}.')
        aliases = expand_to_cypher_alias(self.identproperties, *self.parents+self.relidentproperties+self.dummyrelvars)
        dct = expand_to_cypher_dict(self.dummy, self.propvar, self.identproperties, *self.parents+self.relidentproperties+self.dummyrelvars)
        query = f"""
        CALL apoc.lock.nodes({self.parents}) // let's lock ahead this time
        {optional_match}
        call apoc.do.when({self.dummy} IS NULL, 
                "WITH {aliases} 
                {create}
                SET {self.out} += ${self.propvar}
                RETURN {self.out}, {on_create_rel_returns}",   // created
            "RETURN ${self.dummy} as {self.out}, {on_match_rel_returns}",  // matched 
            {{ {dct} }}) yield value as {self.child_holder}
        WITH *, {rel_expansion}
        """
        return dedent(query)

    @property
    def on_match(self):  # remember, we are in a call context
        query = ''
        for i, (r, rprops, colliding_keys) in enumerate(zip(self.relvars+[self.out], self.relpropsvars+[self.propvar], self.colliding_rel_keys+[self.colliding_keys])):
            if self.collision_manager == 'track&flag':
                if r != self.out:  # handled by the base class above
                    query += dedent(f"""
                        WITH *, [x in apoc.coll.intersection(keys({rprops}), keys(properties({r}))) where ({rprops}[x] is null or {r}[x] is null) or {rprops}[x] <> {r}[x]] as {colliding_keys}
                        CALL apoc.do.when(size({colliding_keys}) > 0, 
                            'WITH $inrel as inrel 
                             MATCH (a)-[inrel]->(b)  
                             CREATE (a)-[c:_Collision]->(b) SET c = $collisions 
                             SET c._dbcreated = $time
                             SET c._reltype = type(inrel)
                             RETURN $time', 
                            'RETURN $time',
                            {{inrel: {r}, collisions: apoc.map.fromLists({colliding_keys}, apoc.map.values({rprops}, {colliding_keys})), time:$time0}}) yield value as _{i}{self.value}
                    """)
            elif self.collision_manager == 'overwrite':
                query += f"\nSET {r} += {rprops}   // overwrite with new colliding properties"
            else:
                query += f"\nSET {r} = apoc.map.merge({rprops}, properties({r}))  // update, keeping the old colliding properties"
        return query

    @property
    def on_create(self):  # remember, we are in a call context
        query = f"SET {self.out}._dbcreated = $time0, {self.out} += {self.propvar}  // setup as standard"
        for r, rprops in zip(self.relvars, self.relpropsvars):
            query += f'\nSET {r}._dbupdated = $time0, {r}._dbcreated = $time0, {r} += {rprops}'
        return query

    @property
    def on_run(self):  # remember, we are in a call context
        query = f"SET {self.out}._dbupdated = time0  // always set updated time"
        for r in self.relvars:
            query += f'\nSET {r}._dbupdated = time0'
        return query

    @property
    def post_merge(self):
        dct = expand_to_cypher_dict(self.out, self.propvar, *self.relvars+self.relpropsvars)
        if len(dct):
            dct += ', '
        dct += 'time0:time0'
        aliases = expand_to_cypher_alias(self.out, self.propvar, *self.relvars+self.relpropsvars)
        return dedent(f"""
        // post merge
        call apoc.do.when({self.dummy} IS NULL,
        "WITH {aliases}\n{self.on_create}\n RETURN $time0",
        "WITH {aliases}\n{self.on_match}\n RETURN $time0",
        {{ {dct} }}) yield value as {self.unnamed}
        {self.on_run}\n""")

    @property
    def collision_record(self):
        return f"WITH $innode as innode CREATE (c:_Collision)-[:COLLIDES]->(innode)"

    @property
    def collision_record_input(self):
        return f"innode: {self.out}"

    def to_cypher(self):
        return super().to_cypher()


class SetVersion(Statement):
    def __init__(self, parents: List[CypherVariable], reltypes: List[str], childlabel: str, child: CypherVariable, childproperties: dict):
        if len(reltypes) != len(parents):
            raise ValueError(f"reltypes must be the same length as parents")
        self.parents = parents
        self.reltypes = reltypes
        self.childlabel = camelcase(childlabel)
        self.childproperties, other_ins = neo4j_dictionary(childproperties)
        self.child = child
        other_ins.append(child)
        super(SetVersion, self).__init__(self.parents+other_ins, [])

    def to_cypher(self):
        matches = ', '.join([f'({p})-[:{r}]->(c:{self.childlabel} {self.childproperties})' for p, r in zip(self.parents, self.reltypes)])
        query = [
            f"WITH * CALL {{",
                f"\t WITH {','.join(map(str, self.parents))}, {self.child}",
                f"\t OPTIONAL MATCH {matches}"
                f"\t WHERE id(c) <> id({self.child})",
                f"\t WITH {self.child}, max(c.version) as maxversion",
                f"\t SET {self.child}.version = coalesce({self.child}['version'], maxversion + 1, 0)",
                f"\t RETURN {self.child}['version']",
            f"}}"
        ]
        return '\n'.join(query)


def match_node(labels, properties):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = MatchNode(labels, properties)
    query.add_statement(statement)
    return statement.out


def match_relationship(parent, child, reltype, properties):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = MatchRelationship(parent, child, reltype, properties)
    query.add_statement(statement)
    return statement.out


def match_pattern_node(labels: List[str], properties: Dict[str, Union[str, int, float, CypherVariable]] = None,
                       parents: List[CypherVariable] = None, children: List[CypherVariable] = None):
    query = CypherQuery.get_context()  # type: CypherQuery
    if properties is None:
        properties = {}
    if parents is None:
        parents = []
    if children is None:
        children = []
    statement = MatchPatternNode(labels, properties, parents, children)
    query.add_statement(statement)
    return statement.out


def merge_single_node(labels, identproperties, properties, collision_manager='track&flag'):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = MergeNode(labels, identproperties, properties, collision_manager)
    query.add_statement(statement)
    return statement.out


def merge_relationship(parent, child, reltype, identproperties, properties, collision_manager='track&flag'):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = MergeRelationship(parent, child, reltype, identproperties, properties, collision_manager)
    query.add_statement(statement)
    return statement.out


def merge_dependent_node(labels, identproperties, properties, parents, reltypes, relidentproperties, relproperties,
                         collision_manager='track&flag'):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = MergeDependentNode(labels, identproperties, properties, parents, reltypes, relidentproperties, relproperties,
                                   collision_manager)
    query.add_statement(statement)
    return statement.outnode


def set_version(parents, reltypes, childlabel, child, childproperties):
    query = CypherQuery.get_context()  # type: CypherQuery
    statement = SetVersion(parents, reltypes, childlabel, child, childproperties)
    query.add_statement(statement)


def merge_node(labels, identproperties, properties=None,
               parents: Dict[CypherVariable, Union[Tuple[str, Optional[Dict], Optional[Dict]], str]] = None,
               versioned_label=None,
               versioned_properties=None,
               collision_manager='track&flag') -> CypherVariable:
    if properties is None:
        properties = {}
    if parents is None:
        parents = {}
    parent_list = []
    reltype_list = []
    relidentproperties_list = []
    relproperties_list = []
    for parent, reldata in parents.items():
        if isinstance(reldata, str):
            reldata = [reldata]
        parent_list.append(parent)
        reltype_list.append(reldata[0])
        if len(reldata) > 1:
            relidentproperties_list.append(reldata[1])
        else:
            relidentproperties_list.append({})
        if len(reldata) > 2:
            relproperties_list.append(reldata[2])
        else:
            relproperties_list.append({})
    if len(parents):
        node = merge_dependent_node(labels, identproperties, properties, parent_list,
                                    reltype_list, relidentproperties_list, relproperties_list,
                                    collision_manager)
    else:
        node = merge_single_node(labels, identproperties, properties, collision_manager)
    if versioned_label is not None:
        if versioned_properties is None:
            versioned_properties = {}
        set_version(parent_list, reltype_list, versioned_label, node, versioned_properties)
    return node



from functools import reduce
from operator import and_
from typing import Union

import networkx as nx
import py2neo
from astropy.table import Table

from weaveio.basequery.common import FrozenQuery
from weaveio.basequery.tree import Branch
from weaveio.writequery import CypherVariable, CypherQuery



class Dissociated(FrozenQuery):
    def __init__(self, handler, branch: Branch, variable: CypherVariable, parent: 'FrozenQuery' = None):
        super().__init__(handler, branch, parent)
        self.variable = variable

    def _filter_by_boolean(self, boolean_filter: 'FrozenQuery'):
        new = self._make_filtered_branch(boolean_filter)
        return self.__class__(self.handler, new, self.variable, self)

    def __getitem__(self, boolean_filter: 'Dissociated'):
        if not isinstance(boolean_filter, Dissociated):
            raise TypeError(f"Factors may only be filtered with `[]` using a boolean filter")
        return self._filter_by_boolean(boolean_filter)

    def _apply_func(self, string, other: 'Dissociated' = None):
        if other is not None:
            if isinstance(other, Dissociated):  # now we have to align first
                aligned = self.branch.align(other.branch)
                inputs = {
                    'x': aligned.action.transformed_variables.get(self.variable, self.variable),
                    'y': aligned.action.transformed_variables.get(other.variable, other.variable)
                }
            elif isinstance(other, FrozenQuery):
                raise TypeError(f"Cannot compare types {self.__class__} and {other.__class__}")
            else:
                aligned = self.branch.add_data(other)
                inputs = {'x': self.variable, 'y': aligned.current_variables[0]}
        else: # only one variable
            inputs = {'x': self.variable}
            aligned = self.branch
        newbranch = aligned.operate(string, **inputs)
        return Dissociated(self.handler, newbranch, newbranch.current_variables[-1], self)

    def __add__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_func('{x} + {y}', other)

    def __sub__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_func('{x} - {y}', other)

    def __mul__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_func('{x} * {y}', other)

    def __truediv__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_func('{x} / {y}', other)

    def __pow__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_func('{x} ** {y}', other)

    __radd__ = __add__
    __rsub__ = __sub__
    __rmul__ = __mul__
    __rtruediv__ = __truediv__
    __rpow__ = __pow__

    def __neg__(self) -> 'Dissociated':
        return self._apply_func('-{x}')

    def __abs__(self) -> 'Dissociated':
        return self._apply_func('abs({x})')

    def __eq__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_func('{x} = {y}', other)

    def __ne__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_func('{x} <> {y}', other)

    def __lt__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_func('{x} < {y}', other)

    def __le__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_func('{x} <= {y}', other)

    def __gt__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_func('{x} > {y}', other)

    def __ge__(self, other: Union['Dissociated', int, float]) -> 'Dissociated':
        return self._apply_func('{x} >= {y}', other)

    def __and__(self, other: 'Dissociated') -> 'Dissociated':
        return self._apply_func('({x} AND {y})', other)

    def __or__(self, other: 'Dissociated') -> 'Dissociated':
        return self._apply_func('({x} OR {y})', other)

    def _post_process(self, result: py2neo.Cursor, squeeze: bool = True) -> Table:
        df = result.to_data_frame()
        assert len(df.columns) == 1
        vs = df.iloc[:, 0].values
        if squeeze and len(vs) == 1:
            return vs[0]
        return vs

    def _prepare_query(self) -> CypherQuery:
        with super()._prepare_query() as query:
            return query.returns(self.variable)


import pytest
import py2neo


def test_update_version_on_first_instance(database: py2neo.Graph, procedure_tag: str):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        
        WITH specs, child
        // So now there is a multimerged thing
        CALL custom.version{procedure_tag}(specs, child, ['A', 'B'], 'version') YIELD version
        WITH collect(version) as v  // just to finish here
        MATCH (child:MyLabel)
        RETURN child.version
        """
    t = database.run(q).to_table()
    assert len(t) == 1
    assert t[0][0] == 0


def test_dont_update_version_on_duplicate_instance(database: py2neo.Graph, procedure_tag: str):
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        CALL custom.version{procedure_tag}(specs, child, ['A', 'B'], 'version') YIELD version
        
        WITH specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        CALL custom.version{procedure_tag}(specs, child, ['A', 'B'], 'version') YIELD version

        WITH collect(version) as v  // just to finish here
        MATCH (child:MyLabel)
        RETURN child.version
        """
    t = database.run(q).to_table()
    assert len(t) == 1
    assert t[0][0] == 0


def test_difference_in_unversioned_rels_provokes_version_increment(database: py2neo.Graph, procedure_tag: str):
    """
    1. Create two sets of parent->child based on A and B type nodes.
    2. Version on A-type nodes only

    When the B nodes differ but the A nodes do not (this test), there should be 2 different versions
    When the A nodes differ but the B nodes do not, there should be 1 version = 0 per child

    """
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        CALL custom.version{procedure_tag}(specs, child, ['A'], 'version') YIELD version

        WITH specs
        MERGE (b2: B {{id: 2}})
        WITH [[b2, 'brel', {{order: 1}}]]+specs as specs  // add another b but dont version on B
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        CALL custom.version{procedure_tag}(specs, child, ['A'], 'version') YIELD version

        WITH collect(version) as v
        MATCH (child: MyLabel)
        MATCH (p)-->(child)
        WITH child, count(p) as cnt
        RETURN child.version, cnt
        ORDER BY cnt
        """
    t = database.run(q).to_table()
    assert len(t) == 2
    assert t[0][0] == 0
    assert t[1][0] == 1


def test_difference_in_versioned_rels_doesnt_provoke_version_increment(database: py2neo.Graph, procedure_tag: str):
    """
    1. Create two sets of parent->child based on A and B type nodes.
    2. Version on A-type nodes only

    When the B nodes differ but the A nodes do not (above test), there should be 2 different versions
    When the A nodes differ but the B nodes do not (this test), there should be 1 version = 0 per child

    """
    q = f"""
        UNWIND range(0, 100) as i
        MERGE (a1: A {{id: i}})
        WITH collect(a1) as input_anodes

        MERGE (b1: B {{id: 1}})

        with input_anodes, [b1] as input_bnodes, ['green'] as input_bnames
        CALL apoc.lock.nodes(input_anodes)
        CALL apoc.lock.nodes(input_bnodes)

        UNWIND RANGE(0, SIZE(input_anodes) - 1) AS ai
        WITH [input_anodes[ai], 'arel', {{order: ai}}] AS arow, input_bnodes, input_bnames
        WITH collect(arow) as anodes, input_bnodes, input_bnames

        UNWIND RANGE(0, SIZE(input_bnodes) - 1) AS bi
        WITH [input_bnodes[bi], 'brel', {{order: bi, name: input_bnames[bi]}}] AS brow, anodes
        WITH collect(brow) as bnodes, anodes
        WITH *, bnodes+anodes as specification

        WITH specification as specs
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        CALL custom.version{procedure_tag}(specs, child, ['A'], 'version') YIELD version

        WITH specs
        MERGE (a2: A {{id: 10000}})
        WITH [[a2, 'arel', {{order: 1}}]]+specs as specs  // add another b but dont version on B
        CALL custom.multimerge{procedure_tag}(specs, ['MyLabel'], {{}}, {{}}, {{}}) YIELD child
        CALL custom.version{procedure_tag}(specs, child, ['A'], 'version') YIELD version

        WITH collect(version) as v
        MATCH (child: MyLabel)
        MATCH (p)-->(child)
        WITH child, count(p) as cnt
        RETURN child.version, cnt
        ORDER BY cnt
        """
    t = database.run(q).to_table()
    assert len(t) == 2
    assert t[0][0] == 0
    assert t[1][0] == 0


from .base import CypherQuery, CypherData, CypherVariable
from .merging import match_node, merge_node, match_relationship, merge_relationship, set_version, \
    match_pattern_node
from .actions import unwind, collect, groupby
from .statements import Unwind, Collection


from copy import deepcopy
from typing import Tuple, Dict, Any

import py2neo

from weaveio.basequery.parse_tree import parse, write_tree, branch2query
from weaveio.basequery.tree import Branch
from weaveio.writequery import CypherQuery


class AmbiguousPathError(Exception):
    pass


class NotYetImplementedError(NotImplementedError):
    pass


class UnexpectedResult(Exception):
    pass


class FrozenQuery:
    executable = True

    def __init__(self, handler, branch: Branch, parent: 'FrozenQuery' = None):
        self.handler = handler
        self.branch = branch
        self.parent = parent
        self.data = self.handler.data
        self.string = ''

    def _make_filtered_branch(self, boolean_filter: 'FrozenQuery'):
        collected = self.branch.collect([boolean_filter.branch], [])
        return collected.filter('{x}', x=collected.action.transformed_variables[boolean_filter.variable])

    def _filter_by_boolean(self, boolean_filter):
        new = self._make_filtered_branch(boolean_filter)
        return self.__class__(self.handler, new, self)

    def _traverse_frozenquery_stages(self):
        query = self
        yield query
        while query.parent is not None:
            query = query.parent
            yield query

    def _prepare_branch(self) -> Branch:
        """Override to allow custom edits to the branch before execution"""
        return self.branch

    def _prepare_query(self) -> CypherQuery:
        """Override to allow custom edits to the CypherQuery object after the branch is finalised"""
        branch = self._prepare_branch()
        query = branch2query(branch)
        return query

    def _prepare_cypher(self) -> Tuple[str, Dict[str, Any]]:
        """Override to allow custom edits to the actual cypher query text"""
        query = self._prepare_query()
        query.remove_variable_names()
        cypher, params = query.render_query()
        return cypher, params

    def _execute_query(self, limit=None):
        """Override to allow custom edits as to how the cypher text is run"""
        if not self.executable:
            raise TypeError(f"{self.__class__} may not be executed as queries in their own right")
        cypher, params = self._prepare_cypher()
        if limit is not None:
            cypher += f'\nLIMIT {limit}'
        return self.data.graph.execute(cypher, **params)

    def _post_process(self, result: py2neo.Cursor, squeeze: bool = True):
        """Override to turn a py2neo neo4j result object into something that the user wants"""
        raise NotImplementedError

    def __call__(self, limit=None, squeeze=True):
        """Prepare and execute the query contained by this frozen object"""
        result = self._execute_query(limit=limit)
        return self._post_process(result, squeeze)

    def __repr__(self):
        return f'{self.parent}{self.string}'


from collections import defaultdict
from typing import List, Union, Type, Tuple

import py2neo

from .common import FrozenQuery, AmbiguousPathError
from .dissociated import Dissociated
from .factor import SingleFactorFrozenQuery, TableFactorFrozenQuery
from .tree import Branch
from ..hierarchy import Hierarchy, Multiple, One2One
from ..writequery import CypherVariable


GET_PRODUCT = "[({{h}})<-[p:product {{{{name: '{name}'}}}}]-(hdu: HDU) | [hdu.sourcefile, hdu.extn, p.index, p.column_name]]"
GET_FACTOR = "{{h}}.{name}"
GET_FACTOR_FORCE_PLURAL = f"[{GET_FACTOR}]"

class HierarchyFrozenQuery(FrozenQuery):
    def __getitem__(self, item):
        raise NotImplementedError

    def __getattr__(self, item):
        raise NotImplementedError

    def _get_factor(self, item, plural):
        raise NotImplementedError(f"Getting singular factors is not supported for {self.__class__.__name__}")

    def _get_hierarchy(self, item, plural):
        raise NotImplementedError(f"Getting singular hierarchies is not supported for {self.__class__.__name__}")

    def _get_table_factor(self, item):
        raise NotImplementedError(f"Getting table factors is not supported for {self.__class__.__name__}")

    def _filter_by_identifiers(self, items):
        raise NotImplementedError(f"Filtering by multiple identifiers is not supported for {self.__class__.__name__}")

    def _filter_by_identifier(self, item):
        raise NotImplementedError(f"Filtering by an identifier is not supported for {self.__class__.__name__}")

    def _filter_by_boolean(self, condition):
        raise NotImplementedError(f"Filtering by a boolean condition is not supported for {self.__class__.__name__}")


class HeterogeneousHierarchyFrozenQuery(HierarchyFrozenQuery):
    """
    The start point for building queries. e.g. `data`
    Available data calls:
        single factor `data.runids` (but always plural)
        single hierarchy `data.runs` (but always plural)
    """
    executable = False

    def __repr__(self):
        return f'query("{self.data.rootdir}/")'

    def _get_hierarchy(self, hierarchy_name, plural) -> 'DefiniteHierarchyFrozenQuery':
        paths, hiers, startbase, endbase = self.handler.paths2hierarchy(hierarchy_name, plural=plural)
        new = self.branch.handler.begin(endbase.__name__)
        return DefiniteHierarchyFrozenQuery(self.handler, new, endbase, new.current_hierarchy, [], self)

    def _get_factor(self, factor_name, plural):
        factor_name = self.data.singular_name(factor_name)
        pathdict, base, is_product = self.handler.paths2factor(factor_name, plural=plural)
        begin = self.branch.handler.begin(base.__name__)
        if is_product:
            func = GET_PRODUCT.format(name=factor_name)
        else:
            func = GET_FACTOR.format(name=factor_name)
        new = begin.operate(func, h=begin.current_hierarchy)
        return SingleFactorFrozenQuery(self.handler, new, factor_name, new.current_variables[0],
                                       plural, is_product, self)

    def __getattr__(self, item):
        if item in self.data.plural_factors:
            return self._get_factor(item, plural=True)
        elif item in self.data.singular_factors:
            raise AmbiguousPathError(f"Cannot return a single factor from a heterogeneous dataset")
        elif item in self.data.singular_hierarchies:
            raise AmbiguousPathError(f"Cannot return a singular hierarchy without filtering first")
        else:
            name = self.data.singular_name(item)
            return self._get_hierarchy(name, plural=True)


class DefiniteHierarchyFrozenQuery(HierarchyFrozenQuery):
    """
    The template class for hierarchy classes that are not heterogeneous i.e. they have a defined hierarchy type
    The start point for building queries. E.g. `data.obs, data.runs.exposure`
    It can have ids or not.
    Available data calls:
        single factor `data.exposures.runids`
        single hierarchy `data.runs.exposure`
        table of factors `data.runs[['runid', 'expmjd']]`
        filter by id `data.runs[11234]`
        filter by condition `data.runs[data.runs.runid > 0]`
    """
    def __init__(self, handler, branch: Branch, hierarchy_type: Type[Hierarchy], hierarchy_variable: CypherVariable,
                 identifiers: List, parent: 'FrozenQuery'):
        super().__init__(handler, branch, parent)
        self.hierarchy_type = hierarchy_type
        self.hierarchy_variable = hierarchy_variable
        self.identifiers = identifiers
        self.string = f"{self.hierarchy_type.singular_name}"

    def _filter_by_boolean(self, boolean_filter: 'FrozenQuery'):
        new = self._make_filtered_branch(boolean_filter)
        return self.__class__(self.handler, new, self.hierarchy_type, self.hierarchy_variable, self.identifiers, self)

    def _prepare_query(self):
        """Add a hierarchy node return statement"""
        query = super(DefiniteHierarchyFrozenQuery, self)._prepare_query()
        with query:
            query.returns(self.branch.find_hierarchies()[-1])
        return query

    def _process_result_row(self, row, nodetype):
        node = row[0]
        if node is None:
            return None
        inputs = {}
        for f in nodetype.factors:
            inputs[f] = node[f]
        if nodetype.idname is not None:
            inputs[nodetype.idname] = node[nodetype.idname]
        base_query = self.handler.hierarchy_from_neo4j_identity(nodetype, node.identity)
        for p in nodetype.parents:
            if isinstance(p, One2One):
                inputs[p.singular_name] = getattr(base_query, p.plural_name)
            elif isinstance(p, Multiple):
                inputs[p.plural_name] = getattr(base_query, p.plural_name)
            else:
                try:
                    inputs[p.singular_name] = getattr(base_query, p.singular_name)
                except AmbiguousPathError:
                    inputs[p.singular_name] = getattr(base_query, p.plural_name)  # this should not have to be done
        h = nodetype(**inputs, do_not_create=True)
        h.add_parent_query(base_query)
        h.add_parent_data(self.handler.data)
        return h

    def _post_process(self, result: py2neo.Cursor, squeeze: bool = True):
        result = result.to_table()
        if len(result) == 1 and result[0] is None:
            return []
        results = []
        for row in result:
            h = self._process_result_row(row, self.hierarchy_type)
            results.append(h)
        if len(results) == 1 and squeeze:
            return results[0]
        return results

    def _get_hierarchy(self, name, plural):
        pathlist, endlist, starthier, endhier = self.handler.paths2hierarchy(name, plural=plural, start=self.hierarchy_type)
        new = self.branch.traverse(*pathlist)
        return DefiniteHierarchyFrozenQuery(self.handler, new, endhier, new.current_hierarchy, [], self)

    def _get_factor_query(self, names: Union[List[str], str], plurals: Union[List[bool], bool]) -> Tuple[Branch, List[CypherVariable], List[bool]]:
        """
        Return the query branch, variables of a list of factor/product names
        We do this by grouping into the containing hierarchies and traversing each branch before collapsing
        returns:
                branch: The new branch object to continue the query with
                variables: The list of CypherVariable that contains the query result
                is_product: The list of True if the factor is a product
        """
        # TODO: tidy this all up
        if not isinstance(names, (list, tuple)):
            names = [names]
        if not isinstance(plurals, (list, tuple)):
            plurals = [plurals]
        names = [self.data.singular_name(name) for name in names]
        local = []
        remote = defaultdict(list)
        remote_paths = {}
        for name, plural in zip(names, plurals):
            pathsdict, basehier, is_product = self.handler.paths2factor(name, plural, self.hierarchy_type)
            if basehier == self.hierarchy_type:
                local.append((name, plural, is_product))
            else:
                remote[(basehier, plural)].append((name, is_product))
                remote_paths[(basehier, plural)] = {path for pathset in pathsdict.values() for path in pathset}

        variables = {}
        is_products = {}
        branch = self.branch

        for (basehier, plural), factor_product_tuples in remote.items():
            paths = remote_paths[(basehier, plural)]
            travel = branch.traverse(*paths)
            funcs = []
            for name, is_product in factor_product_tuples:
                if is_product:
                    func = GET_PRODUCT.format(name=name)
                else:
                    func = GET_FACTOR.format(name=name)
                funcs.append(func)
            operate = travel.operate(*funcs, h=travel.current_hierarchy)
            if plural:
                branch = branch.collect([], [operate])
            else:
                branch = branch.collect([operate], [])
            for v, (name, is_product) in zip(operate.current_variables, factor_product_tuples):
                variables[(name, plural)] = branch.action.transformed_variables[v]
                is_products[(name, plural)] = is_product

        if len(local):
            funcs = []
            for name, plural, is_product in local:
                if is_product:
                    func = GET_PRODUCT.format(name=name)
                elif plural:
                    func = GET_FACTOR_FORCE_PLURAL.format(name=name)
                else:
                    func = GET_FACTOR.format(name=name)
                funcs.append(func)
            branch = branch.operate(*funcs, h=self.hierarchy_variable)
            for v, (k, plural, is_product) in zip(branch.action.output_variables, local):
                variables[(k, plural)] = v
                is_products[(k, plural)] = is_product

        # now propagate variables forward
        values = [variables[(name, plural)] for name, plural in zip(names, plurals)]
        variables = branch.get_variables(values)
        return branch, variables, [is_products[(name, plural)] for name, plural in zip(names, plurals)]

    def _get_factor(self, name, plural):
        branch, factor_variables, is_products = self._get_factor_query([name], [plural])
        return SingleFactorFrozenQuery(self.handler, branch, name, factor_variables[0], plural,
                                       is_products[0], self)

    def _get_factor_table_query(self, item) -> TableFactorFrozenQuery:
        """
        __getitem__ is for returning factors and ids
        There are three types of getitem input values:
        List: [[a, b]], where labelled table-like rows are output
        Tuple: [a, b], where a list of unlabelled dictionaries are output
        str: [a], where a single value is returned

        In all three cases, you still need to specify plural or singular forms.
        This allows you to have a row of n dimensional heterogeneous data.
        returns query and the labels (if any) for the table
        """
        if isinstance(item, tuple):  # return without headers
            return_keys = list(item)
            keys = list(item)
        elif isinstance(item, list):
            keys = item
            return_keys = item
        elif item is None:
            raise TypeError("item must be of type list, tuple, or str")
        else:
            raise KeyError(f"Unknown item {item} for `{self}`")
        plurals = [not self.data.is_singular_name(i) for i in item]
        branch, factor_variables, is_products = self._get_factor_query(keys, plurals)
        return TableFactorFrozenQuery(self.handler, branch, keys, factor_variables, plurals, is_products, return_keys, self.parent)

    def _filter_by_identifiers(self, identifiers: List[Union[str, int, float]]) -> 'DefiniteHierarchyFrozenQuery':
        idname = self.hierarchy_type.idname
        new = self.branch.add_data(identifiers)
        identifiers_var = new.current_variables[0]
        branch = new.filter('{h}.' + idname + ' in {identifiers}', h=self.hierarchy_variable, identifiers=identifiers_var)
        return DefiniteHierarchyFrozenQuery(self.handler, branch, self.hierarchy_type, self.hierarchy_variable, identifiers, self)

    def _filter_by_identifier(self, identifier: Union[str, int, float]) -> 'DefiniteHierarchyFrozenQuery':
        idname = self.hierarchy_type.idname
        new = self.branch.add_data(identifier)
        identifier_var = new.current_variables[0]
        branch = new.filter('{h}.' + idname + ' = {identifier}', h=self.hierarchy_variable, identifier=identifier_var)
        return DefiniteHierarchyFrozenQuery(self.handler, branch, self.hierarchy_type, self.hierarchy_variable, [identifier], self)

    def __getitem__(self, item):
        """
        If there is a scalar, then return a SingleFactorFrozenQuery otherwise assume its a request for a table
        The order of resolution is:
            vector hierarchy/
        """
        if isinstance(item, Dissociated):
            return self._filter_by_boolean(item)
        if isinstance(item, (list, tuple)):
            if all(map(self.data.is_valid_name, item)):
                return self._get_factor_table_query(item)
            elif any(map(self.data.is_valid_name, item)):
                raise KeyError(f"You cannot mix IDs and names in a __getitem__ call")
            else:
                return self._filter_by_identifiers(item)
        if not self.data.is_valid_name(item):
            return self._filter_by_identifier(item)  # then assume its an ID
        else:
            return getattr(self, item)

    def __getattr__(self, item):
        plural = self.data.is_plural_name(item)
        factor = self.data.is_factor_name(item)
        exists = self.data.is_singular_name(item) or plural
        if not exists:
            raise AttributeError(f"{self} has no attribute {item}")
        if factor:
            return self._get_factor(item, plural=plural)
        else:
            return self._get_hierarchy(item, plural=plural)



from collections import defaultdict
from functools import reduce
from operator import xor
from typing import List, Dict, Optional

import networkx as nx
from networkx import OrderedDiGraph

from weaveio.basequery.actions import Action, EntryPoint, StartingPoint, DataReference, Traversal, Collection, Aggregation, Operation, Filter, shared_branch_without_filters, Results, TraversalPath, ScalarAlignment, DifferentLevelAlignment
from weaveio.writequery.base import CypherVariable, CypherData


class BranchHandler:
    def __init__(self):
        self.graph = OrderedDiGraph()
        self.class_counter = defaultdict(int)
        self.entry = self.new(EntryPoint(), [], [], None, [], [], [])
        self.data_objects = {}

    def new(self, action: Action, accessible_parents: List['Branch'], inaccessible_parents: List['Branch'],
            current_hierarchy: Optional[CypherVariable], current_variables: List[CypherVariable],
            variables: List[CypherVariable], hierarchies: List[CypherVariable], name: str = None):
        parents = accessible_parents + inaccessible_parents
        parent_set = set(parents)
        successors = {s for p in parents for s in self.graph.successors(p) if set(self.graph.predecessors(s)) == parent_set}
        candidates = {s for s in successors if s.action == action}
        assert len(candidates) <= 1
        if candidates:
            return successors.pop()
        if name is None:
            self.class_counter[action.__class__] += 1
            name = action.__class__.__name__ + str(self.class_counter[action.__class__])
        instance = Branch(self, action, accessible_parents, inaccessible_parents, current_hierarchy, current_variables,
                          variables=variables, hierarchies=hierarchies, name=name)
        attrs = {}
        if action.shape is not None:
            attrs['shape'] = action.shape
        self.graph.add_node(instance, action=action, name=name, **attrs)
        for parent in parents:
            self.graph.add_edge(parent, instance, accessible=parent in accessible_parents)
        self.current_hierarchy = current_hierarchy
        return instance

    def begin(self, *labels):
        action = StartingPoint(*labels)
        return self.new(action, [self.entry], [], action.hierarchy, [action.hierarchy], [], [action.hierarchy])

    def relevant_graph(self, branch):
        return nx.subgraph_view(self.graph, lambda n: nx.has_path(self.graph, n, branch) or n is branch)

    def deepest_common_ancestor(self, *branches: 'Branch'):
        common = set()
        for i, branch in enumerate(branches):
            ancestors = set(nx.algorithms.dag.ancestors(self.graph, branch))
            if i == 0:
                common = ancestors
            else:
                common &= ancestors
        if not len(common):
            return None
        distances = [(sum(nx.shortest_path_length(self.graph, ancestor, b) for b in branches), ancestor) for ancestor in common]
        return distances[distances.index(min(distances, key=lambda x: x[0]))][1]

    def _align_scalar(self, shared: 'Branch', vector: 'Branch', scalar: 'Branch'):
        scalar_collected = shared.collect([scalar], [])
        action = ScalarAlignment(vector, [scalar_collected])
        action.transformed_variables = scalar_collected.action.transformed_variables
        return self.new(action, [vector], [scalar_collected], None, action.outs, [], [])

    def _align_different_level(self, shared: 'Branch', branch1: 'Branch', branch2: 'Branch'):
        """
        1. collect the vector back to the shared level
        2. coalesce the scalar back to the shared level
        3. unwind the vector
        """
        collected1 = shared.collect([], [branch1])
        collected2 = shared.collect([], [branch2])
        action = DifferentLevelAlignment(shared, [collected1, collected2])
        transformed_variables = {}
        shared_variables = {v: v for v in shared.find_variables()}
        for v in branch1.find_variables():
            if v not in shared_variables and not isinstance(v, CypherData):
                transformed_variables[v] = action.transformed_variables[collected1.action.transformed_variables[v]]
        for v in branch2.find_variables():
            if v not in shared_variables and not isinstance(v, CypherData):
                try:
                    transformed_variables[v] = action.transformed_variables[collected2.action.transformed_variables[v]]
                except KeyError:
                    pass
        transformed_variables.update(shared_variables)
        action.transformed_variables = transformed_variables
        return self.new(action, [shared], [collected1, collected2], None, action.outs, [], [])



def plot(graph, fname):
    from networkx.drawing.nx_agraph import to_agraph
    A = to_agraph(graph)
    A.layout('dot')
    A.draw(fname)


class Branch:
    def __init__(self, handler: BranchHandler, action: Action,
                 accessible_parents: List['Branch'], inaccessible_parents: List['Branch'],
                 current_hierarchy: Optional[CypherVariable], current_variables: List[CypherVariable],
                 hierarchies: List[CypherVariable],
                 variables: List[CypherVariable], name: str = None):
        """
        A branch is an object that represents all the Actions (query statements) in a query.
        It contains both a node (Action) and references to all actions (other Branches) preceeding it (parents).
        If a branch is created in the same way more than once, only one object is actually instantiated, this is to optimize performance
        when writing to cypher query language. The user shouldn't care about this quirk.
        :param handler: The handler object which oversees query uniqueness for a particular dataset.
        :param action: The action that this branch will execute at the end. These are generally (but not always) Cypher statements.
        :param parents: The branches that must be executed before this one (i.e. the dependencies of this branch)
        """
        self.handler = handler
        self.action = action
        self.parents = accessible_parents + inaccessible_parents
        self.accessible_parents = accessible_parents
        self.inaccessible_parents = inaccessible_parents
        self.current_hierarchy = current_hierarchy
        self.current_variables = current_variables
        self.name = name
        self.hierarchies = hierarchies
        self.variables = variables
        self._relevant_graph = None
        self._accessible_graph = None
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = reduce(xor, map(hash, self.parents + [self.current_hierarchy, tuple(self.current_variables), self.action, self.handler]))
        return self._hash

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        return self.handler == other.handler and self.action == other.action and set(self.parents) == set(other.parents) \
               and self.current_hierarchy == other.current_hierarchy and set(self.current_variables) == set(other.current_variables)

    def __repr__(self):
        return f'<Branch {self.name}: {str(self.action)}>'

    @property
    def relevant_graph(self):
        if self._relevant_graph is None:
            self._relevant_graph = self.handler.relevant_graph(self)
        return self._relevant_graph

    @property
    def accessible_graph(self):
        if self._accessible_graph is None:
            g = nx.subgraph_view(self.relevant_graph, filter_edge=lambda a, b: self.relevant_graph.edges[(a, b)]['accessible'])
            self._accessible_graph = nx.subgraph_view(g, filter_node=lambda x: nx.has_path(g, x, self))
        return self._accessible_graph

    def iterdown(self, graph=None, start=None):
        if graph is None:
            graph = self.relevant_graph
        if start is None:
            start = self.handler.entry
        return nx.algorithms.traversal.dfs_tree(graph, start)

    def find_hierarchies(self):
        hierarchies = []
        for branch in self.iterdown(self.accessible_graph):
            if branch.current_hierarchy is not None:
                hierarchies.append(branch.current_hierarchy)
        return hierarchies

    def find_variables(self):
        variables = []
        for branch in self.iterdown(self.accessible_graph):
            variables += branch.current_variables
        return variables

    def find_hierarchy_branches(self, entry=False):
        branches = []
        for branch in self.iterdown(self.accessible_graph):
            if branch.current_hierarchy is not None or (branch is self.handler.entry and entry):
                branches.append(branch)
        return branches

    def get_variables(self, variables: List[CypherVariable]):
        squeeze = False
        if not isinstance(variables, (list, tuple)):
            variables = [variables]
            squeeze = True
        accessible_variables = self.find_variables()
        for branch in self.iterdown(self.relevant_graph):
            for i, variable in enumerate(variables):
                if variable in accessible_variables:
                    pass
                if variable in branch.current_variables:
                    pass
                else:
                    try:
                        variables[i] = branch.action.transformed_variables[variable]
                    except (AttributeError, KeyError):
                        pass
        if squeeze and len(variables) == 1:
            return variables[0]
        return variables


    def add_data(self, *data) -> 'Branch':
        action = DataReference(*data)
        return self.handler.new(action, [self], [], current_hierarchy=None, current_variables=action.input_variables,
                         variables=self.variables + action.input_variables, hierarchies=self.hierarchies)

    def traverse(self, *paths: TraversalPath) -> 'Branch':
        """
        Extend the branch from the most recent hierarchy to a new hierarchy(s) by walking along the paths
        If more than one path is given then we take the union of all the resultant nodes.
        """
        action = Traversal(self.find_hierarchies()[-1], *paths)
        return self.handler.new(action, [self], [], current_hierarchy=action.out, current_variables=[action.out],
                                variables=self.variables, hierarchies=self.hierarchies+[action.out])

    # def align(self, branch: 'Branch') -> 'Branch':
    #     """
    #     Join branches into one, keeping the highest cardinality.
    #     This is used to to directly compare arrays:
    #         * ob1.runs == ob2.runs  (unequal sizes are zipped up together)
    #         * ob1.runs == run1  (array to single comparisons are left as is)
    #         * run1 == run2  (single to single comparisons are left as is)
    #     zip ups and unwinds take place relative to the branch's shared ancestor
    #     """
    #     shared = join_branches(self, branch)
    #     a = shared.collect([self], [])
    #     b = shared.collect([branch], [])
    #     action = Alignment(a, b)
    #     return self.handler.new(action, [a], [b], None, current_variables=action.outs,
    #                             variables=a.variables, hierarchies=a.hierarchies)

    def align(self, branch: 'Branch') -> 'Branch':
        """
        Aligning a scalar (1 of them is at the level of the shared hierarchy):
        =======================================================================================
            To align things like: data.obs.targets.ra == max(data.obs.runids, wrt=data.obs),
            we find the shared parent (data.obs)
            for each:
                if hierarchy level is above the shared parent, collect the results
                if hierarchy level is at the shared parent, do nothing
                if hierarchy level is below the shared parent, raise error
            we are now at the level of the shared parent
            zip up and unwind the variables together so they are aligned (using the length of variable which was collected)

        Aligning two different branches (neither of them are at the level of shared hierarchy):
        =======================================================================================
            To align things like data.obs[all(data.obs.targets[x].runs == data.obs.targets[y].runs)]
                i.e. this thing:  data.obs.targets[x].runs == data.obs.targets[y].runs
            we find the shared parent (data.obs)
            for each:
                collect the results
            we are now at the level of the shared parent
            zip up and unwind the variables together so they are aligned (using the length of variable which was collected)

        Aligning on the same branch:
        =======================================================================================
            To align things like:
                1) data.obs.runs == data.obs.runs
                    they share the exact same tree, so `.align` does nothing
                2) data.obs.runs == max(data.obs.runs.target.ra, wrt=data.obs.runs)
                    the share the same tree but since the right-hand-side splits off and aggregates back,
                    we do nothing but return the split off branch
        """
        shared = shared_branch_without_filters(self, branch)
        branch_is_scalar = branch.find_hierarchy_branches(True)[-1] is shared
        self_is_scalar = self.find_hierarchy_branches(True)[-1] is shared
        if branch_is_scalar and self_is_scalar:
            return self.handler._align_different_level(shared, self, branch)
        if branch_is_scalar:
            return self.handler._align_scalar(shared, self, branch)
        elif self_is_scalar:
            return self.handler._align_scalar(shared, branch, self)
        else:
            return self.handler._align_different_level(shared, self, branch)



    def collect(self, singular: List['Branch'], multiple: List['Branch']) -> 'Branch':
        """
        Join branches into one, reducing the cardinality to this branch.
        `singular` contains branches that will be coalesced (i.e. only the first result is taken)
        `multiple` contains branches that will be collected (i.e. all results are presented in a list)
        This is used in predicate filters:
            ob1.runs[any(ob1.runs.l1singlespectra.snr > 10)]
            0. branch `ob1.runs` is created
            1. branch `ob1.runs.l1singlespectra.snr > 10` is created
            2. branch `ob1.runs.l1singlespectra.snr > 10` is collected with respect to `ob1.runs`
            3. A filter is applied on the collection at the `ob1.runs` level
        After a collection, only
        """
        action = Collection(self, singular, multiple)
        variables = action.outsingle_variables + action.outmultiple_variables
        hierarchies = action.outsingle_hierarchies + action.outmultiple_hierarchies
        return self.handler.new(action, [self], singular + multiple, None, variables + hierarchies,
                                variables=self.variables + variables, hierarchies=self.hierarchies + hierarchies)

    def aggregate(self, string_function: str, variable: CypherVariable, branch: 'Branch', remove_infs: bool, namehint: str) -> 'Branch':
        action = Aggregation(string_function, variable, branch, self, remove_infs, namehint)
        return self.handler.new(action, [self], [branch], None, [action.target], [action.target], [])

    def operate(self, *string_functions, namehint=None, **inputs: CypherVariable) -> 'Branch':
        """
        Adds a new variable to the namespace
        e.g. y = x*2 uses extant variable x to define a new variable y which is then subsequently accessible
        """
        # missing = [k for k, v in inputs.items() if getattr(v, 'parent', v) not in self.variables + self.hierarchies]
        # if missing:
        #     raise ValueError(f"inputs {missing} are not in scope for {self}")
        op = Operation(*string_functions, namehint=namehint, **inputs)
        return self.handler.new(op, [self], [], None, op.output_variables,
                                variables=self.variables + op.output_variables, hierarchies=self.hierarchies)

    def filter(self, logical_string, **boolean_variables: CypherVariable) -> 'Branch':
        """
        Reduces the cardinality of the branch by using a WHERE clause.
        .filter can only use available variables
        """
        # missing = [k for k, v in boolean_variables.items() if getattr(v, 'parent', v) not in self.variables + self.hierarchies]
        # if missing:
        #     raise ValueError(f"inputs {missing} are not in scope for {self}")
        action = Filter(logical_string, **boolean_variables)
        return self.handler.new(action, [self], [], None, [],
                                variables=self.variables, hierarchies=self.hierarchies)

    def results(self, branch_attributes: Dict['Branch', List[CypherVariable]]):
        """
        Returns the rows of results.
        All other branches which are not this branch will have their results collected
        """
        branch_attributes = {k: v if isinstance(v, (list, tuple)) else [v] for k, v in branch_attributes.items()}
        action = Results(branch_attributes)
        return self.handler.new(action, [self], [], None, [], self.variables, self.hierarchies)



from datetime import datetime

import py2neo
import pytest
from py2neo.wiring import WireError

from weaveio.graph import Graph


@pytest.fixture(scope='module')
def procedure_tag():
    return str(hash(datetime.now())).replace('-', '')[:5]


@pytest.fixture(scope='function')
def write_database(procedure_tag) -> Graph:
    try:
        graph = Graph(port=7687, host='host.docker.internal')
        assert graph.neograph.name == 'testweaveiodonotuse', "I will not run tests on this database as a safety measure"
        graph.neograph.run('MATCH (n) DETACH DELETE n')
        graph.neograph.run('CALL apoc.schema.assert({},{},true) YIELD label, key RETURN *')
        graph.neograph.run('call db.clearQueryCaches')
        return graph
    except (AssertionError, WireError):
        pytest.xfail("unsupported configuration of testing database")




import py2neo
import pytest
from py2neo.wiring import WireError

from weaveio.basequery.query_objects import Path
from weaveio.basequery.tests.example_structures.one2one import MyDataOne2One


@pytest.fixture(scope='session')
def workdir(tmpdir_factory):
    d = tmpdir_factory.mktemp("data")
    for i in range(5):
        fname = Path(str(d.join(f'{i}.fits')))
        with open(str(fname), 'w') as file:
            file.write('')
    return d


@pytest.fixture
def data_one2one(workdir):
    return MyDataOne2One(workdir, port=None)  # set to None for safety


@pytest.fixture(scope='module')
def database_one2one(workdir):
    try:
        data = MyDataOne2One(workdir, port=7687, write=True)
        assert data.graph.neograph.name == 'testweaveiodonotuse', "I will not run tests on this database as a safety measure"
        data.graph.neograph.run('MATCH (n) DETACH DELETE n')
        data.graph.neograph.run('CALL apoc.schema.assert({},{},true) YIELD label, key RETURN *')
    except (AssertionError, WireError):
        pytest.xfail("unsupported configuration of testing database")
    else:
        data.read_directory()
        data.validate()
        return data


from pathlib import Path
from typing import Union, List, Tuple, Dict

from astropy.io import fits
from astropy.io.fits.hdu.base import _BaseHDU

from weaveio.graph import Graph
from weaveio.hierarchy import Hierarchy


class File(Hierarchy):
    is_template = True
    idname = 'fname'
    match_pattern = '*.file'
    hdus = {}
    produces = []
    recommended_batchsize = None

    def open(self):
        try:
            return fits.open(self.data.rootdir / self.fname)
        except AttributeError:
            return fits.open(self.fname)

    def __init__(self, fname, **kwargs):
        super().__init__(tables=None, fname=str(fname), **kwargs)

    @classmethod
    def get_batches(cls, path, batch_size):
        if batch_size is None:
            return [slice(None, None)]
        n = cls.length(path)
        return (slice(i, i + batch_size) for i in range(0, n, batch_size))

    @classmethod
    def match_file(cls, directory: Union[Path, str], fname: Union[Path, str], graph: Graph):
        """Returns True if the given fname in a given directory can be read by this class of file hierarchy object"""
        fname = Path(fname)
        return fname.match(cls.match_pattern)

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None) -> 'File':
        raise NotImplementedError

    @classmethod
    def read_hdus(cls, directory: Union[Path, str], fname: Union[Path, str],
                  **hierarchies: Union[Hierarchy, List[Hierarchy]]) -> Tuple[Dict[str,'HDU'], 'File', List[_BaseHDU]]:
        path = Path(directory) / Path(fname)
        file = cls(fname, **hierarchies)
        hdus = [i for i in fits.open(path)]
        if len(hdus) != len(cls.hdus):
            raise TypeError(f"Class {cls} asserts there are {len(cls.hdus)} HDUs ({list(cls.hdus.keys())})"
                            f" whereas {path} has {len(hdus)} ({[i.name for i in hdus]})")
        hduinstances = {}
        for i, ((hduname, hduclass), hdu) in enumerate(zip(cls.hdus.items(), hdus)):
            hduinstances[hduname] = hduclass.from_hdu(hduname, hdu, i, file)
        return hduinstances, file, hdus

    def read_product(self, product_name):
        self.build_index()
        return getattr(self, f'read_{product_name}')()


class HDU(Hierarchy):
    is_template = True
    parents = [File]
    factors = ['sourcefile', 'extn', 'name']
    identifier_builder = ['sourcefile', 'extn', 'name']
    binaries = ['header', 'data']
    concatenation_constants = None

    @classmethod
    def _from_hdu(cls, hdu):
        return {}

    @classmethod
    def from_hdu(cls, name, hdu, extn, file):
        input_dict = cls._from_hdu(hdu)
        input_dict[cls.parents[0].singular_name] = file
        input_dict['extn'] = extn
        input_dict['sourcefile'] = file.fname
        input_dict['name'] = name
        if cls.concatenation_constants is not None:
            if len(cls.concatenation_constants):
                for c in cls.concatenation_constants:
                    if c not in input_dict:
                        input_dict[c] = hdu.header[c]
                input_dict['concatenation_constants'] = cls.concatenation_constants
        return cls(**input_dict)


class PrimaryHDU(HDU):
    is_template = True
    binaries = ['header']
    concatenation_constants = []


class BaseDataHDU(HDU):
    is_template = True
    concatenation_constants = ['ncols']
    factors = HDU.factors + ['nrows', 'ncols']


class TableHDU(BaseDataHDU):
    is_template = True
    concatenation_constants = ['columns']

    @classmethod
    def _from_hdu(cls, hdu):
        input_dict = BaseDataHDU._from_hdu(hdu)
        if hdu.data is None:
            input_dict['columns'] = []
            input_dict['nrows'] = 0
            input_dict['ncols'] = 0
        else:
            colnames = [str(i) for i in hdu.data.names]
            input_dict['columns'] = colnames
            input_dict['nrows'], input_dict['ncols'] = hdu.data.shape[0], len(colnames)
        return input_dict


class BinaryHDU(BaseDataHDU):
    is_template = True

    @classmethod
    def _from_hdu(cls, hdu):
        input_dict = BaseDataHDU._from_hdu(hdu)
        if hdu.data is None:
            input_dict['nrows'], input_dict['ncols'] = 0, 0
        else:
            input_dict['nrows'], input_dict['ncols'] = hdu.data.shape
        return input_dict


class SpectralBlockHDU(BinaryHDU):
    is_template = True
    concatenation_constants = ['naxis1', 'naxis2']


class SpectralRowableBlock(BinaryHDU):
    is_template = True
    concatenation_constants = ['naxis1', 'crval1', 'cunit1', 'cd1_1']



from datetime import datetime

import py2neo
import pytest
from py2neo.wiring import WireError

from weaveio.neo4jquery.procedures.make_procedure import get_all_procedures


@pytest.fixture(scope='module')
def procedure_tag():
    return str(hash(datetime.now())).replace('-', '')[:5]


@pytest.fixture(scope='function')
def database(procedure_tag) -> py2neo.Graph:
    try:
        graph = py2neo.Graph(port=7687, host='host.docker.internal')
        assert graph.name == 'testweaveiodonotuse', "I will not run tests on this database as a safety measure"
        graph.run('MATCH (n) DETACH DELETE n')
        for text in get_all_procedures('write', procedure_tag):
            graph.run(text)
        return graph
    except (AssertionError, WireError):
        pytest.xfail("unsupported configuration of testing database")




from pathlib import Path
from typing import List

import py2neo
from astropy.io import fits
from astropy.table import Table, Column
import pandas as pd
import numpy as np
from tqdm import tqdm

from weaveio.basequery.common import FrozenQuery
from weaveio.basequery.dissociated import Dissociated
from weaveio.basequery.tree import Branch
from weaveio.writequery import CypherVariable, CypherQuery


def replace_with_data(row, files):
    hdus = files[row['sourcefile']]
    hdu = hdus[row['extn']]
    if not pd.isnull(row['column_name']):
        table = Table(hdu.data)
        table = table[row['column_name']]
    else:
        table = hdu.data
    if not pd.isnull(row['index']):
        return table[row['index']]
    return table


class FactorFrozenQuery(Dissociated):
    def __init__(self, handler, branch: Branch, factors: List[str], factor_variables: List[CypherVariable],
                 plurals: List[bool], is_products: List[bool], parent: FrozenQuery = None):
        super().__init__(handler, branch, factor_variables[0], parent)
        self.factors = factors
        self.factor_variables = factor_variables
        self.plurals = plurals
        self.is_products = is_products

    def _filter_by_boolean(self, boolean_filter: 'Dissociated'):
        new = self._make_filtered_branch(boolean_filter)
        return self.__class__(self.handler, new, self.factors, self.factor_variables, self.plurals,
                              self.is_products, self)

    def _prepare_query(self) -> CypherQuery:
        with super()._prepare_query() as query:
            return query.returns(*self.factor_variables)

    def __repr__(self):
        if isinstance(self.factors, tuple):
            factors = f'{self.factors}'
        else:
            factors = f'[{self.factors}]'
        return f'{self.parent}{factors}'

    def _parse_products(self, df: pd.DataFrame):
        """
        Take a pandas dataframe and replace the structure of ['sourcefile', 'extn', 'index', 'column_name']
        with the actual data
        """
        product_columns = []
        for colname, is_product in zip(df.columns, self.is_products):
            if is_product:
                df[colname] = df[colname].apply(lambda x: [None]*4 if len(x) == 0 else x[0])
                split = df[colname].apply(pd.Series)
                split.columns = ['sourcefile', 'extn', 'index', 'column_name']
                split.index.name = 'rown'
                split['colname'] = colname
                split.set_index('colname', append=True, inplace=True)
                product_columns.append(split)
        if len(product_columns):
            sourcefiles = {}
            concat = pd.concat(product_columns)
            for fname in tqdm(concat.sourcefile.drop_duplicates(), desc='reading fits files'):
                path = Path(self.handler.data.rootdir) / fname
                sourcefiles[fname] = fits.open(path)
            data = concat.apply(replace_with_data, files=sourcefiles, axis='columns').sort_index(level=0)  # type: pd.Series
            for name, group in data.groupby('colname'):
                df[name] = group.values
        return df

    def _post_process(self, result: py2neo.Cursor, squeeze: bool = True) -> Table:
        df = self._parse_products(result.to_data_frame())
        # replace lists with arrays
        for c in df.columns:
            if df.dtypes[c] == 'O' and not isinstance(df[c].iloc[0], str):
                df[c] = df[c].apply(np.asarray)
        table = Table.from_pandas(df)
        for colname, plural, is_product in zip(df.columns, self.plurals, self.is_products):
            if plural or is_product:
                if df.dtypes[colname] == 'O':
                    shapes = set(map(np.shape, df[colname]))
                    if len(shapes) == 1:  # all the same length
                        table[colname] = Column(np.stack(df[colname].values), name=colname, shape=shapes.pop(), length=len(df))
        if len(table) == 1 and squeeze:
            table = table[0]
        if len(table.colnames) == 1 and squeeze:
            table = table[table.colnames[0]]
        return table


class SingleFactorFrozenQuery(FactorFrozenQuery):
    def __init__(self, handler, branch: Branch, factor: str, factor_variable: CypherVariable,
                 plural: bool, is_product: bool, parent: FrozenQuery = None):
        super().__init__(handler, branch, [factor], [factor_variable], [plural], [is_product], parent)


class TableFactorFrozenQuery(FactorFrozenQuery):
    """
    A matrix of different factors against different hierarchy instances
    This is only possible if the hierarchies each have only one of the factors
    """
    def __init__(self, handler, branch, factors, factor_variables, plurals, is_products,
                 return_keys: List[str] = None, parent: 'FrozenQuery' = None):
        super().__init__(handler, branch, factors, factor_variables, plurals, is_products, parent)
        self.return_keys = return_keys

    def _prepare_query(self) -> CypherQuery:
        with super()._prepare_query() as query:
            variables = {k: v for k, v in zip(self.return_keys, self.factor_variables)}
            return query.returns(**variables)

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __getitem__(self, item):
        if isinstance(item, str):
            try:
                i = self.factors.index(item)
            except ValueError:
                raise KeyError(f"{item} is not a factor contained within {self}. Only {self.factors} are accessible.")
            else:
                return SingleFactorFrozenQuery(self.handler, self.branch, item, self.factor_variables[i],
                                               self.plurals[i], self.is_products[i], self)
        else:
            return super(TableFactorFrozenQuery, self).__getitem__(item)



import os
from pathlib import Path

HERE = Path(os.path.dirname(os.path.abspath(__file__)))

def read_procedure_text(cql_fname):
    with open(HERE / Path(cql_fname), 'r') as f:
        cypherlines = [l.strip() for l in f.readlines()]
    for i ,line in enumerate(cypherlines):
        if all(s == '/' for s in line):
            break
    else:
        raise ValueError('CQL header not found')
    header = cypherlines[:i]
    params = []
    returns = []
    requires = []
    description = 'no description given'
    for l in header:
        if '//' in l:
            l = l.split('//')[1].strip()
            if l.startswith('param:'):
                params.append([i.strip() for i in l[len('param:'):].split('=>')])
            if l.startswith('returns:'):
                returns.append([i.strip() for i in l[len('returns:'):].split('=>')])
            if l.startswith('description:'):
                description = l[len('description:'):].strip()
            if l.startswith('requires:'):
                requires = [x.strip() for x in l[len('requires:'):].strip().split(',')]
    query = '\n'.join(cypherlines[i+1:]).replace("'", "\\'")
    return query, returns, params, description, requires


def make_procedure(name, cql_fname, readwrite, tag=''):
    query, returns, params, description, requirements = read_procedure_text(cql_fname)
    for req in requirements:
        query = query.replace(req, f'custom.{req}{tag}')
    return f"CALL apoc.custom.asProcedure('{name}{tag}', '{query}', '{readwrite}', {returns}, {params}, '{description}')"


def get_all_procedures(mode, tag=''):
    procedures = []
    for f in HERE.glob('*.cql'):
        name = f.name.replace('.cql', '')
        procedures.append(make_procedure(name, str(f), mode, tag))
    return procedures



import numpy as np
import pytest
from hypothesis import given, settings, note, Verbosity, HealthCheck, example, reproduce_failure, assume, reject, event
from hypothesis import strategies as st

from .test_hypothesis_strategies import create_node_from_node, node_strategy, Node, is_equal, neo4jproperty_strategy, rel_strategy, create_rel_from_rel
from .. import CypherQuery, merge_node
from ..actions import custom
from ..base import Varname, CypherData
from ..merging import PropertyOverlapError, are_different
from ...graph import Graph, is_null

SYSTEM_PROPERTIES = ['_dbupdated', '_dbcreated']

def assert_is_equal(a, b):
    if not is_equal(a, b):  # can handle nans
        assert a == b  # cant handle nans, but has a nicer output


@pytest.mark.parametrize('nantype', [np.nan, float('nan')])
def test_nans_convert_to_none(write_database, nantype):
    r = write_database.execute('return {a: $nan}', nan=nantype).to_table()[0][0]
    assert_is_equal({'a': None}, r)


def test_nones(write_database):
    r = write_database.execute('return {a: $none}', none=None).to_table()[0][0]
    assert_is_equal({'a': None}, r)


def _test_difference(write_database, a, b):
    with CypherQuery() as query:
        _a = CypherData(a, 'a')
        _b = CypherData(b, 'b')
        diff = custom(lambda ai, bi, out: f'WITH *, {are_different(ai, bi)} as {out}', [_a, _b], ['diff'])[0]
        query.returns(diff=diff, a=_a, b=_b)
    cypher, params = query.render_query()
    note(cypher)
    are_actually_different, neoa, neob = write_database.execute(cypher, **params).to_table()[0]
    note(f'a was input as {a} and output as {neoa}')
    note(f'b was input as {b} and output as {neob}')
    return are_actually_different

@pytest.mark.diff
@settings(deadline=None, max_examples=200, verbosity=Verbosity.normal)
@example(x=np.nan)
@example(x=[np.nan])
@example(x=None)
@example(x=[None])
@given(x=neo4jproperty_strategy)
def test_not_different(write_database, x):
    are_actually_different = _test_difference(write_database, x, x)
    assert are_actually_different == False


@pytest.mark.diff
@settings(deadline=5000, max_examples=200, verbosity=Verbosity.normal)
@example(x=np.nan, y=1)
@example(x=[1, np.nan], y=[1, 2])
@example(x=np.nan, y=[np.nan])
@example(x=None, y=1)
@example(x=[1, None], y=[1, 2])
@example(x=None, y=[None])
@given(x=neo4jproperty_strategy, y=neo4jproperty_strategy)
def test_different(write_database, x, y):
    assume(not is_equal(x, y))
    are_actually_different = _test_difference(write_database, x, y)
    assert are_actually_different == True


def test_raw_null_returns_none(write_database):
    assert write_database.execute('RETURN null').to_table()[0][0] is None


def test_input_none_returns_none(write_database):
    assert write_database.execute('RETURN $n', n=None).to_table()[0][0] == np.inf


def test_input_nan_returns_none(write_database):
    assert write_database.execute('RETURN $n', n=np.nan).to_table()[0][0] == np.inf


def test_input_list_nan_returns_none(write_database):
    assert write_database.execute('RETURN $n', n=[np.nan, 1.0]).to_table()[0][0] == [np.inf, 1]


def assert_overlapping_properties(write_database, time, node1, node2, collision_manager, exclude=None):
    """
    Once node1 and node2 have been identified as the same, assert that the set properties are correct
    If collisions occur, the behaviour is set by collision_manager (overwrite, ignore, or track&flag)
        overwrite will replace colliding keys
        ignore will do nothing with colliding keys
        track&flag will not change the node, but will place the colliding keys into a collision node attached to it
    New keys will always be added to the node
    Unmentioned keys (in matched node but not in proposed properties) will always be ignored
    If the entire properties dict is different, then this policy has the effect of just adding more properties.
    """
    if exclude is None:
        exclude = []
    exclude = ' AND '.join([f'NOT n:{e}' for e in exclude])
    if len(exclude):
        exclude = f' AND {exclude} '
    collision = None
    same_keys = set(node1.properties.keys()) & set(node2.properties.keys())
    colliding_keys = [k for k in same_keys if not is_equal(node1.properties[k], node2.properties[k])]
    added_keys = set(node2.properties.keys()) - set(node1.properties.keys())

    new = {k: node2.properties[k] for k in added_keys}
    expected = node1.allproperties.copy()
    expected.update(new)  # always add new
    expected.update({'_dbcreated': time, '_dbupdated': time})

    if collision_manager == 'ignore':
        pass
    elif collision_manager == 'overwrite':
        expected.update(node2.properties)
    elif collision_manager == 'track&flag':
        if len(colliding_keys):
            collision = {k: node2.properties[k] for k in colliding_keys}
            collision.update({'_dbcreated': time})
    else:
        raise ValueError(f"Unknown collision_manager {collision_manager}")

    nodes = write_database.execute(f'MATCH (n) WHERE NOT n:_Collision {exclude}\n'
                                   'OPTIONAL MATCH (n)<-[:COLLIDES]-(c:_Collision)\n'
                                   'RETURN {properties: properties(n), labels: labels(n)} as n, {properties: properties(c), labels: labels(c)} as c').to_table()
    assert len(nodes) == 1
    node, collision_node = nodes[0]
    assert is_equal(node['properties'], expected)
    if collision is None:
        assert is_equal(collision_node, {'properties': None, 'labels': None})
    else:
        assert is_equal(collision_node['properties'], collision)


def assert_overlapping_rel_properties(write_database, time, parent, child, rel1, rel2, collision_manager):
    collision = None
    same_keys = set(rel1.properties.keys()) & set(rel2.properties.keys())
    colliding_keys = [k for k in same_keys if not is_equal(rel1.properties[k], rel2.properties[k])]
    added_keys = set(rel2.properties.keys()) - set(rel1.properties.keys())

    new = {k: rel2.properties[k] for k in added_keys}
    expected = rel1.allproperties.copy()
    expected.update(new)  # always add new
    expected.update({'_dbcreated': time, '_dbupdated': time})

    if collision_manager == 'ignore':
        pass
    elif collision_manager == 'overwrite':
        expected.update(rel2.properties)
    elif collision_manager == 'track&flag':
        if len(colliding_keys):
            collision = {k: rel2.properties[k] for k in colliding_keys}
            collision.update({'_dbcreated': time})
    else:
        raise ValueError(f"Unknown collision_manager {collision_manager}")
    alabels = ':'.join(parent.labels)
    blabels = ':'.join(child.labels)
    aprops = parent.allproperties.copy()
    bprops = child.allproperties.copy()
    aprops.update({'_dbcreated': time, '_dbupdated': time})
    bprops.update({'_dbcreated': time, '_dbupdated': time})
    query = f'MATCH (a:{alabels})-[r]->(b:{blabels}) WHERE properties(a) = $aprops AND properties(b) = $bprops AND NOT r:_Collision\n'\
                                   'OPTIONAL MATCH (a)-[c:_Collision]->(b)\n'\
                                   'RETURN {properties: properties(r), type: type(r)} as r, {properties: properties(c), type: type(c)} as c'
    note(query)
    note(aprops)
    note(bprops)
    rels = write_database.execute(query, aprops=aprops, bprops=bprops).to_table()
    assert len(rels) == 1
    rel, collision_rel = rels[0]
    assert is_equal(rel['properties'], expected)
    if collision is None:
        assert is_equal(collision_rel, {'properties': None, 'type': None})
    else:
        assert is_equal(collision_rel['properties'], collision)


def assert_only_these_nodes_exist(write_database, time, *nodes, exclude=None):
    if exclude is None:
        exclude = []
    exclude = ' AND '.join([f'NOT n:{e}' for e in exclude])
    if len(exclude):
        exclude = f' AND {exclude} '
    for node in nodes:
        number = sum(node == n for n in nodes)
        labels = ':'.join(node.labels)
        props = node.allproperties.copy()
        props.update({'_dbcreated': time, '_dbupdated': time})
        props = {k: v for k, v in props.items()}
        query = f'MATCH (n:{labels}) WHERE properties(n) = $props AND size(labels(n)) = {len(node.labels)} {exclude} RETURN n'
        note(query)
        result = write_database.execute(query, props=props).to_table()
        assert len(result) == number
    assert write_database.execute(f'MATCH (n) RETURN count(n)').to_table()[0][0] == len(nodes)


def assert_merge_node_tests(node1, node2, collision_manager, write_database, time):
    has_equivalent_labels = all(l in node1.labels for l in node2.labels)
    has_equivalent_ident = all(node1.identproperties.get(k, None) == v for k,v in node2.identproperties.items())
    matched = has_equivalent_labels and has_equivalent_ident
    if matched:
        assert_overlapping_properties(write_database, time, node1, node2, collision_manager)
    else:
        assert_only_these_nodes_exist(write_database, time, node1, node2)

@settings(max_examples=1000, deadline=None, suppress_health_check=[HealthCheck.data_too_large],
          # verbosity=Verbosity.verbose,
          print_blob=True)
@given(node=node_strategy, sampler=st.data(),
collision_manager=st.sampled_from(['track&flag', 'overwrite', 'ignore']),
different_labels=st.sampled_from(['crop', 'entire', False, 'extend', 'crop&extend']),
different_properties=st.sampled_from([False, 'entirekeys', 'crop', 'addkeys', 'overwritekeys']),
different_identproperties=st.sampled_from([False, 'entirekeys', 'crop', 'addkeys', 'overwritekeys']))
def test_node_pair(node, sampler, different_properties, different_identproperties, different_labels, collision_manager, write_database):
    write_database.neograph.run('MATCH (n) DETACH DELETE n')
    write_database.neograph.run('CALL apoc.schema.assert({},{},true) YIELD label, key RETURN *')
    write_database.neograph.run('call db.clearQueryCaches')
    node1 = node
    node2 = create_node_from_node(node, sampler, different_labels, different_properties, different_identproperties)
    note(node1)
    note(node2)
    assume(not any(any(k in n.identproperties for k in n.properties) for n in [node1, node2]))  # no overlaps

    with CypherQuery() as query:
        a = merge_node(node1.labels, node1.identproperties, node1.properties, collision_manager=collision_manager)
        b = merge_node(node2.labels, node2.identproperties, node2.properties, collision_manager=collision_manager)
    cypher, params = query.render_query()
    time = write_database.execute(cypher, **params).to_table()[0][0]
    assert_merge_node_tests(node1, node2, collision_manager, write_database, time)


@pytest.mark.parametrize('collision_manager', ['track&flag', 'overwrite', 'ignore'])
def test_node_pair_with_parents_different_rels_properties(write_database: Graph, collision_manager):
    """
    create the nparent nodes
    """
    write_database.neograph.run('MATCH (n) DETACH DELETE n')
    write_database.neograph.run('CALL apoc.schema.assert({},{},true) YIELD label, key RETURN *')
    write_database.neograph.run('call db.clearQueryCaches')
    with CypherQuery() as query:
        a = merge_node(['A'], {})
        b = merge_node(['B'], {})
        for i in range(3):
            merge_node(['C'], {'id': 1}, {'c': 1}, parents={a: [f'arel', {'a': 1}, {'b': i}],
                                                            b: ['brel', {'a': 2}, {'b': i}]},
                       collision_manager=collision_manager)
    cypher, params = query.render_query()
    time = write_database.execute(cypher, **params).to_table()[0][0]
    assert len(write_database.execute('MATCH (c:_Collision) RETURN c').to_table()) == 0
    collision_result = write_database.execute('MATCH ()-[c:_Collision]->() RETURN c ORDER BY c.b, c._reltype').to_series()
    if collision_manager == 'track&flag':
        assert len(collision_result) == 4  # one is create, the other 2 collide, times by 2 and so there should be 4
        assert [i['b'] for i in collision_result.values] == [1, 1, 2, 2]  #  two collisions, two properties in each
        assert [i['_reltype'] for i in collision_result.values] == ['arel', 'brel', 'arel', 'brel']  #  two collisions, two properties in each
    else:
        assert len(collision_result) == 0
    result = write_database.execute('MATCH (a:A)-[ar:arel]->(c:C)<-[br:brel]-(b:B) '
                                    'RETURN properties(a) as a, properties(b) as b, properties(c) as c, '
                                    'properties(ar) as ar, properties(br) as br').to_data_frame()
    assert len(result) == 1
    result = result.iloc[0]
    assert result['c'] == {'c': 1, 'id': 1, '_dbupdated': time, '_dbcreated': time}
    if collision_manager == 'overwrite':
        assert result['ar'] == {'a': 1, 'b': 2, '_dbupdated': time, '_dbcreated': time}
        assert result['br'] == {'a': 2, 'b': 2, '_dbupdated': time, '_dbcreated': time}
    else:
        assert result['ar'] == {'a': 1, 'b': 0, '_dbupdated': time, '_dbcreated': time}
        assert result['br'] == {'a': 2, 'b': 0, '_dbupdated': time, '_dbcreated': time}


@pytest.mark.parametrize('collision_manager', ['track&flag', 'overwrite', 'ignore'])
def test_node_pair_with_parents_different_parents(write_database: Graph, collision_manager):
    """
    create the nparent nodes
    """
    write_database.neograph.run('MATCH (n) DETACH DELETE n')
    write_database.neograph.run('CALL apoc.schema.assert({},{},true) YIELD label, key RETURN *')
    write_database.neograph.run('call db.clearQueryCaches')
    with CypherQuery() as query:
        a = merge_node(['A'], {})
        b1 = merge_node(['B'], {'id': 1})
        b2 = merge_node(['B'], {'id': 2})
        merge_node(['C'], {'id': 1}, {'c': 1}, parents={a: [f'arel', {'a': 1}, {'b': 1}],
                                                        b1: ['brel', {'a': 2}, {'b': 1}]},
                   collision_manager=collision_manager)
        merge_node(['C'], {'id': 1}, {'c': 1}, parents={a: [f'arel', {'a': 1}, {'b': 1}],
                                                        b2: ['brel', {'a': 2}, {'b': 1}]},
                   collision_manager=collision_manager)
    cypher, params = query.render_query()
    time = write_database.execute(cypher, **params).to_table()[0][0]
    assert len(write_database.execute('MATCH (c:_Collision) RETURN c').to_table()) == 0
    collision_result = write_database.execute('MATCH ()-[c:_Collision]->() RETURN c ORDER BY c.b, c._reltype').to_series()
    assert len(collision_result) == 0
    result = write_database.execute('MATCH (a:A)-[ar:arel]->(c:C)<-[br:brel]-(b:B) '
                                    'RETURN properties(a) as a, properties(b) as b, properties(c) as c, '
                                    'properties(ar) as ar, properties(br) as br').to_data_frame()
    assert len(result) == 2


@pytest.mark.parametrize('collision_manager', ['track&flag', 'overwrite', 'ignore'])
def test_node_pair_with_parents_different_childprops(write_database: Graph, collision_manager):
    """
    create the nparent nodes
    """
    write_database.neograph.run('MATCH (n) DETACH DELETE n')
    write_database.neograph.run('CALL apoc.schema.assert({},{},true) YIELD label, key RETURN *')
    write_database.neograph.run('call db.clearQueryCaches')
    with CypherQuery() as query:
        a = merge_node(['A'], {})
        b1 = merge_node(['B'], {'id': 1})
        random = merge_node(['B'], {'id': 2})
        merge_node(['C'], {'id': 1}, {'c': 1}, parents={a: [f'arel', {'a': 1}, {'b': 1}],
                                                        b1: ['brel', {'a': 2}, {'b': 1}]},
                   collision_manager=collision_manager)
        merge_node(['C'], {'id': 2}, {'c': 1}, parents={a: [f'arel', {'a': 1}, {'b': 1}],
                                                        b1: ['brel', {'a': 2}, {'b': 1}]},
                   collision_manager=collision_manager)
    cypher, params = query.render_query()
    time = write_database.execute(cypher, **params).to_table()[0][0]
    assert len(write_database.execute('MATCH (c:_Collision) RETURN c').to_table()) == 0
    collision_result = write_database.execute('MATCH ()-[c:_Collision]->() RETURN c ORDER BY c.b, c._reltype').to_series()
    assert len(collision_result) == 0
    result = write_database.execute('MATCH (a:A)-[ar:arel]->(c:C)<-[br:brel]-(b:B) '
                                    'RETURN properties(a) as a, properties(b) as b, properties(c) as c, '
                                    'properties(ar) as ar, properties(br) as br').to_data_frame()
    assert len(result) == 2


def assert_merge_dependent_node_tests(node1, node2, parenta1, parentb1, parenta2, parentb2, rela1, relb1, rela2, relb2, ndifferent_parents, collision_manager, write_database, time):
    has_equivalent_labels = all(l in node1.labels for l in node2.labels)
    has_equivalent_ident = all(node1.identproperties.get(k, None) == v for k,v in node2.identproperties.items())
    has_equivalent_parent_labels = all(all(l in old.labels for l in new.labels) for old, new in [[parenta1, parenta2], [parentb1, parentb2]])
    has_equivalent_parent_ident = all(all(old.identproperties.get(k, None) == v for k, v in new.identproperties.items()) for old, new in [[parenta1, parenta2], [parentb1, parentb2]])
    has_equivalent_rel_labels = all(old.reltype == new.reltype for old, new in [[rela1, rela2], [relb1, relb2]])
    has_equivalent_rel_ident = all(all(old.identproperties.get(k, None) == v for k, v in new.identproperties.items()) for old, new in [[rela1, rela2], [relb1, relb2]])

    matched = has_equivalent_labels and has_equivalent_ident and has_equivalent_parent_labels and has_equivalent_parent_ident\
              and has_equivalent_rel_labels and has_equivalent_rel_ident
    if matched:
        assert_overlapping_properties(write_database, time, node1, node2, collision_manager, ['ParentA', 'ParentB'])
        assert_overlapping_rel_properties(write_database, time, parenta1, node1, rela1, rela2, collision_manager)
        assert_overlapping_rel_properties(write_database, time, parentb1, node1, relb1, relb2, collision_manager)
    else:
        if ndifferent_parents == 0:
            assert_only_these_nodes_exist(write_database, time, node1, node2, parenta1, parentb1)
        elif ndifferent_parents == 1:
            assert_only_these_nodes_exist(write_database, time, node1, node2, parenta1, parentb1, parenta2)
        elif ndifferent_parents == 2:
            assert_only_these_nodes_exist(write_database, time, node1, node2, parenta1, parentb1, parenta2, parentb2)
        else:
            raise NotImplementedError(f"Unhandled number of different parents {ndifferent_parents}")


@settings(max_examples=1000, deadline=None, suppress_health_check=[HealthCheck.data_too_large, HealthCheck.too_slow],
          print_blob=True)
@given(
    node1=node_strategy, sampler=st.data(), rela=rel_strategy, relb=rel_strategy,
    collision_manager=st.sampled_from(['overwrite', 'ignore', 'track&flag']),
    different_labels=st.sampled_from([False, 'crop', 'entire', 'extend', 'crop&extend']),
    different_properties=st.sampled_from([False, 'entirekeys', 'crop', 'addkeys', 'overwritekeys']),
    different_identproperties=st.sampled_from([False, 'entirekeys', 'crop', 'addkeys', 'overwritekeys']),
    ndifferent_parents=st.sampled_from([0, 1, 2]),
    different_rela_reltype=st.sampled_from([False, 'entire']),  # rels only have one
    different_rela_properties=st.sampled_from([False, 'entirekeys', 'crop', 'addkeys', 'overwritekeys']),
    different_rela_identproperties=st.sampled_from([False, 'entirekeys', 'crop', 'addkeys', 'overwritekeys']),
    different_relb_reltype=st.sampled_from([False, 'entire']),  # rels only have one
    different_relb_properties=st.sampled_from([False, 'entirekeys', 'crop', 'addkeys', 'overwritekeys']),
    different_relb_identproperties=st.sampled_from([False, 'entirekeys', 'crop', 'addkeys', 'overwritekeys']),
)
def test_dependent_node_merge(write_database, node1, rela, relb, sampler, collision_manager, different_labels, different_properties,
                              different_identproperties, ndifferent_parents, different_rela_reltype,
                              different_rela_properties, different_rela_identproperties, different_relb_reltype,
                              different_relb_properties, different_relb_identproperties):

    write_database.neograph.run('MATCH (n) DETACH DELETE n')
    write_database.neograph.run('CALL apoc.schema.assert({},{},true) YIELD label, key RETURN *')
    write_database.neograph.run('call db.clearQueryCaches')

    parenta = Node(['ParentA'], {}, identproperties={'id': 1})  # no props so we dont get extra collisions
    parentb = Node(['ParentB'], {}, identproperties={'id': 1})  # no props so we dont get extra collisions
    parenta2 = Node(['ParentA'], {}, identproperties={'id': 1})  # clone it
    parentb2 = Node(['ParentB'], {}, identproperties={'id': 1})  # clone it
    if ndifferent_parents > 0:
        parenta2.identproperties['id'] = 2
    if ndifferent_parents > 1:
        parentb2.identproperties['id'] = 2

    node2 = create_node_from_node(node1, sampler, different_labels, different_properties, different_identproperties)
    rela2 = create_rel_from_rel(rela, sampler, different_rela_reltype, different_rela_properties, different_rela_identproperties)
    relb2 = create_rel_from_rel(relb, sampler, different_relb_reltype, different_relb_properties, different_relb_identproperties)

    assume(not any(any(k in n.identproperties for k in n.properties) for n in [node1, node2, rela, relb, rela2, relb2, parenta, parentb, parenta2, parentb2]))  # no overlaps
    for pa, pb, ch in [[parenta, parentb, node1], [parenta2, parentb2, node2]]:
        for thing in ['labels', 'identproperties']:
            assume(not is_equal(getattr(pa, thing), getattr(ch, thing)))
            assume(not is_equal(getattr(pb, thing), getattr(ch, thing)))

    note(f'node1 = {node1}')
    note(f'node2 = {node2}')
    note(f'parenta1 = {parenta}')
    note(f'parenta2 = {parenta2}')
    note(f'parentb1 = {parentb}')
    note(f'parentb2 = {parentb2}')
    note(f'rela2 = {rela2}')
    note(f'relb2 = {relb2}')

    with CypherQuery() as query:
        parenta_node = merge_node(parenta.labels, parenta.identproperties, parenta.properties, collision_manager=collision_manager)
        parentb_node = merge_node(parentb.labels, parentb.identproperties, parentb.properties, collision_manager=collision_manager)
        parenta2_node = merge_node(parenta2.labels, parenta2.identproperties, parenta2.properties, collision_manager=collision_manager)
        parentb2_node = merge_node(parentb2.labels, parentb2.identproperties, parentb2.properties, collision_manager=collision_manager)

        child1 = merge_node(node1.labels, node1.identproperties, node1.properties, parents={parenta_node: (rela.reltype, rela.identproperties, rela.properties),
                                                                                           parentb_node: (relb.reltype, relb.identproperties, relb.properties)},
                            collision_manager=collision_manager)
        child2 = merge_node(node2.labels, node2.identproperties, node2.properties, parents={parenta2_node: (rela2.reltype, rela2.identproperties, rela2.properties),
                                                                                            parentb2_node: (relb2.reltype, relb2.identproperties, relb2.properties)},
                            collision_manager=collision_manager)
    cypher, params = query.render_query()
    time = write_database.execute(cypher, **params).to_table()[0][0]

    assert_merge_dependent_node_tests(node1, node2, parenta, parentb, parenta2, parentb2, rela, relb, rela2, relb2, ndifferent_parents, collision_manager, write_database, time)